with Field;
with Status;
with Position; use Position;

package body Field_Control is
   use Positions_List;

   protected body Barrier is
      function Get_Neighbourhood(Player_Position : in Coordinates_T)
                                 return Field.Field_Area_T is
      begin
         return (X_Range_T'Max(X_Range_T'First, Player_Position.X - 2),
                 X_Range_T'Min(X_Range_T'Last, Player_Position.X + 2),
                 Y_Range_T'Max(Y_Range_T'First, Player_Position.Y - 2),
                 Y_Range_T'Min(Y_Range_T'Last, Player_Position.Y + 2));
      end Get_Neighbourhood;

      -- a function to check if a player can access the protocol
      function Check_For_Access(Player_Position : in Coordinates_T)
                                return Boolean is
         -- an area to represent the player neighbourhood (+/- 2 cells)
         Player_Neighbourhood : constant Field.Field_Area_T :=
           Get_Neighbourhood(Player_Position);
         -- a boolean to store if we find a player in the neighbourhood
         Found : Boolean := False;
         -- the cursor to and the value of the current element in the iteration
         Current_Element_Cursor : Positions_List.Cursor := Allowed.First;
         Current_Element : Coordinates_T;
      begin
         -- player can access the protocol only if there are no already allowed
         -- players in his neighbourhood; to check we have to iterate on the
         -- allowed list and return False if we find a position in the neighbourhood
         while Current_Element_Cursor /= Positions_List.No_Element loop
            Current_Element := Positions_List.Element(Current_Element_Cursor);
            -- check if the player neighbourhood contains the current position
            if Field.Contains(Player_Neighbourhood, Current_Element) then
               Found := True;
            else
               -- the current element is not in the player neighbourhood
               -- so we can go on with the next element
               Current_Element_Cursor := Positions_List.Next(Current_Element_Cursor);
            end if;
            exit when Found;
         end loop;
         return not Found;
      end Check_For_Access;

      entry Play(Player_Position : in Coordinates_T)
        when True is
      begin
         -- Time analysis
         if Playing = 0 and Start < Status.Get_Corresponding_Quantum(Clock) then
            TS_Sum := TS_Sum + TS;
            TS_Records := TS_Records + 1;
            if TS > WCET then
               WCET := TS;
            end if;
            Start := Status.Get_Corresponding_Quantum(Clock);
         end if;

         Playing := Playing + 1;

         -- check if the player can access the protocol
         if Check_For_Access(Player_Position) then
            -- yes he can, add his position to the Append list and allow him to
            -- play
            Allowed.Append(Player_Position);
         else
            -- no he can't, requeue him to the Awaiting_Players entry
            requeue Awaiting_Players with abort;
         end if;
      end Play;

      procedure Leave(Player_Position : in Coordinates_T) is
         Position_Cursor : Positions_List.Cursor := Allowed.Find(Player_Position);
      begin
         -- unregister his position from the Allowed list (only if it exists)
         if Allowed.Contains(Player_position) then
            Allowed.Delete(Position_Cursor);
            Playing := Playing - 1;
            TS := Clock - Start;
         end if;

         -- if there's somebody waiting to play open the Awaiting_Players guard
         -- and store the number of players who are potentially allowable
         Potentially_Allowable := Awaiting_Players'Count;
         if Potentially_Allowable > 0 then
            Someone_Finished := True;
         end if;
      end;

      procedure Read_ETs(ACET : out Time_Span; WCET : out Time_Span) is
      begin
         ACET := TS_Sum / Natural'Max(1,TS_Records);
         WCET := Barrier.WCET;
      end;

      entry Awaiting_Players(Player_Position : in Coordinates_T)
        when Someone_Finished is
      begin
         -- decrement the number of potentially allowable players and close
         -- the entry guard if the number reaces zero
         Potentially_Allowable := Potentially_Allowable - 1;
         if Potentially_Allowable = 0 then
            Someone_Finished := False;
         end if;

         -- check if the current awaiting player has now access to the protocol
         if Check_For_Access(Player_Position) then
            -- yes he can, add his position to the Append list and allow him to
            -- play
            Allowed.Append(Player_Position);
         else
            -- no he can't, requeue him again to the Awaiting_Players entry
            requeue Awaiting_Players;
         end if;
      end Awaiting_Players;
   end Barrier;
end Field_Control;
