with Ada.Text_IO;    use Ada.Text_IO;
with Status; use Status;
with Tactic; use Tactic;
with Events; use Events;
with Events.Change_Tactic; use Events.Change_Tactic;
with Billboard_Output; use Billboard_Output;
with Input_Types; use Input_Types;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Distribution_Control; use Distribution_Control;
with Game_Control; use Game_Control;
with Assistant_Referee; use Assistant_Referee;
with Referee; use Referee;
with Input_Sources.Strings; use Input_Sources.Strings;
with DOM.Core; use DOM.Core;
with DOM.Core.Documents; use DOM.Core.Documents;
with DOM.Core.Nodes; use DOM.Core.Nodes;
with DOM.Core.Attrs; use DOM.Core.Attrs;
with DOM.Readers; use DOM.Readers;
with Sax.Readers; use Sax.Readers;
with Unicode.CES.Utf8; use Unicode.CES.Utf8;


--  <description>
--    This package interact with Core to execute the GUI requests
--  </description>
--  <group>latinada</group>
--  <c_version>1.0.0</c_version>
package body Core_Listener is

   procedure Get_Initial_Configuration
     (LocalManagerGUI_connecteds   : in integer;
      VisitorManagerGUI_connecteds : in integer;
      SpectatorGUI_connecteds      : in integer) is

      Current_Configuration : constant Simulation_Configuration_T :=
        Simulation_Configuration.Get;

   begin
      -- check if the xml configuration file is initialized
      if Simulation_Xml_String = Null_Unbounded_String then
         -- configuration not initialized, response with fail
         null;
      else
         -- configuration initialized, reply with the configuration
         Notify_Initial_Configuration
           (LocalManagerGUI_connecteds   => LocalManagerGUI_connecteds,
            VisitorManagerGUI_connecteds   => VisitorManagerGUI_connecteds,
            SpectatorGUI_connecteds => SpectatorGUI_connecteds,
            Current_Status          => Game_State.Get.Status,
            Configuration_Xml       => To_String(Simulation_Xml_String),
            Period_Duration	    => Current_Configuration.Period_Duration,
            Break_Duration	    => Current_Configuration.Break_Duration,
            Substitution_Allowed=> Current_Configuration.Substitutions_Allowed);
      end if;

      --@process request and response with a Remote Procedure Call
   end Get_Initial_Configuration;

   procedure Start_Match
     (Period_Duration                               : in integer;
      Break_Duration                                : in integer;
      Substitutions_Allowed                         : in integer) is

      Success : Boolean;

   begin
      -- try to set the configuration sent by SimulatorGUI
      Simulation_Configuration.Set(
        Simulation_Configuration_T'(
          Period_Duration       => Period_Duration,
          Break_Duration        => Break_Duration,
          Substitutions_Allowed => Substitutions_Allowed),
        Success);
      -- check if the operation successed
      if Success then
         -- success, send OK response
         Notify_Response_Inquiry(Destination    => simulatorGUI,
                                 Reply_Type     => "startMatch",
                                 Reply_Status   => True,
                                 Reply_Numeric_Detail => 0);
      else
         -- fail, send FAIL response
         Notify_Response_Inquiry(Destination    => simulatorGUI,
                                 Reply_Type     => "startMatch",
                                 Reply_Status   => False,
                                 Reply_Numeric_Detail => 0);
      end if;

      --@process request and response with a Remote Procedure Call
   end Start_Match;

   procedure Stop_Match is
   begin
      -- set the new game status; this causes the exit of all players
      Game_State.Abort_Match;
      -- wait for all players
      Game_State.Wait_The_End;
      -- deactive the assistant referee task (resetting its ball out time)
      Assistant_Referee_Task.Reset_Ball_Out_Time;
      -- kill the referee
      Referee_Task.Abort_Match;
      -- response
      Notify_Response_Inquiry(Destination    => simulatorGUI,
                              Reply_Type     => "stopMatch",
                              Reply_Status   => True,
                              Reply_Numeric_Detail => 0);

      --@process request and response with a Remote Procedure Call
   end Stop_Match;

   procedure Start_Second_Half is
   begin
      -- request the end of the half time break to the referee
      Referee_Task.End_Half_Time_Break;
      -- response
      Notify_Response_Inquiry(Destination    => simulatorGUI,
                              Reply_Type     => "startSecondHalf",
                              Reply_Status   => True,
                              Reply_Numeric_Detail => 0);

      --@process request and response with a Remote Procedure Call
   end Start_Second_Half;

   procedure Get_Manager_Configuration
     (Manager_Caller            : in Input_Types.Caller) is
      Team_Xml_Id : Unbounded_String;
      Team_Formation_Id : Natural;
      Team_Pressing_Balance_Value : Natural;
   begin
      -- get the current teams configuration for the caller team
      case Manager_Caller is
         when localManagerGUI =>
            Team_Xml_Id := Teams_Status(Local).Get_Id;
            Team_Formation_Id := Teams_Status(Local).Get_Formation.id;
            Team_Pressing_Balance_Value := Teams_Status(Local).Get_Pressing;
         when visitorManagerGUI =>
            Team_Xml_Id := Teams_Status(Visitor).Get_Id;
            Team_Formation_Id := Teams_Status(Visitor).Get_Formation.Id;
            Team_Pressing_Balance_Value := Teams_Status(Visitor).Get_Pressing;
         when others =>
            -- others values are not possible, do nothing
            null;
      end case;
      -- response
      Notify_Started_Lineup(ManagerGUI        => Manager_Caller,
                            Configuration_Xml => To_String(Teams_Xml_String),
                            Match_Status      => Game_State.Get.Status,
                            Team_Id           => To_String(Team_Xml_Id),
                            Team_Formation    => Team_Formation_Id,
                            Pressing_Level    => Team_Pressing_Balance_Value);
   end Get_Manager_Configuration;

   procedure Save_Team_Configuration
     (Team_Id: in Team_Id_T;
      Configuration : in string) is

      Input : String_Input;
      Reader : Tree_Reader;
      Doc : Document;
      Team_Configuration : Team_Configuration_T;

      Team_Element : DOM.Core.Element;
      Current_Player_Node : Node;
      I : Positive := Player_Number_T'First;

      Reply_Status : Configuration_Code_T;
      Xml_Team_Id : Unbounded_String;
      Formation_Id : Natural;
      Pressing_Level : Integer;

   begin
      -- parse the configuration (sent as XML)
      Set_Public_Id(Input, "Teams Configuration");
      Open(Configuration, Utf8_Encoding, Input);
      Set_Feature(Reader, Validation_Feature, False);
      Set_Feature(Reader, Namespace_Feature, False);
      Parse(Reader, Input);
      Close(Input);
      Doc := Get_Tree(Reader);
      -- root team element
      Team_Element := Get_Element(Doc);
      -- get the team configuration from team element (team id, formation id and
      -- pressing value)
      Team_Configuration.Team_Id :=
        To_Unbounded_String(Value(Get_Named_Item(Attributes(Team_Element),"id")));
      Team_Configuration.Formation_Id :=
        Natural'Value(Value(Get_Named_Item(Attributes(Team_Element),"formation_id")));
      Team_Configuration.Pressing_Balance_Value :=
        Integer'Value(Value(Get_Named_Item(Attributes(Team_Element),"pressing")));
      -- the first player xml DOM node
      Current_Player_Node := First_Child(Team_Element);
      -- iterate on team players
      while Current_Player_Node /= null loop
         -- process only element node
         if Node_Type(Current_Player_Node) = Element_Node then
            -- for each player get the player id, the role index and the boolean
            -- to say if the player is first-string
            Team_Configuration.Players_Configuration(I).Player_Id :=
              To_Unbounded_String(Value(Get_Named_Item(Attributes(Current_Player_Node),"id")));
            Team_Configuration.Players_Configuration(I).Player_Role_Index :=
              Natural'Value(Value(Get_Named_Item(Attributes(Current_Player_Node),"role")));
            if Value(Get_Named_Item(Attributes(Current_Player_Node),"is-first-string")) = "true" then
               Team_Configuration.Players_Configuration(I).Is_First_String := True;
            else
               Team_Configuration.Players_Configuration(I).Is_First_String := False;
            end if;
            I := I + 1;
         end if;
         -- load the next player
         Current_Player_Node := Next_Sibling(Current_Player_Node);
      end loop;
      -- free the memory
      Free(Reader);
      -- set the team configuration
      Teams_Configuration.Set(Team_Id, Team_Configuration, Reply_Status);
      -- response
      declare
         Local_Teams_Configuration : constant Teams_Configuration_T :=
           Teams_Configuration.Get;
      begin

         case Team_Id is
         when Local =>
            -- get the local team configuration
            Xml_Team_Id := Local_Teams_Configuration(Local).Team_Id;
            Formation_Id := Local_Teams_Configuration(Local).Formation_Id;
            Pressing_Level := Local_Teams_Configuration(Local).Pressing_Balance_Value;
            -- send response
            Response_Manager_Save(Destination    => localManagerGUI,
                                  Reply_Type     => "saveTeamConfig",
                                  Reply_Status   => Reply_Status,
                                  Team_Id        => To_String(Xml_Team_Id),
                                  Formation_Id   => Formation_Id,
                                  Pressing_Level => Pressing_Level);
         when Visitor =>
            -- get the visitor team configuration
            Xml_Team_Id := Local_Teams_Configuration(Visitor).Team_Id;
            Formation_Id := Local_Teams_Configuration(Visitor).Formation_Id;
            Pressing_Level := Local_Teams_Configuration(Visitor).Pressing_Balance_Value;
            -- send response
            Response_Manager_Save(Destination    => visitorManagerGUI,
                                  Reply_Type     => "saveTeamConfig",
                                  Reply_Status   => Reply_Status,
                                  Team_Id        => To_String(Xml_Team_Id),
                                  Formation_Id   => Formation_Id,
                                  Pressing_Level => Pressing_Level);
         end case;
      end;
   end Save_Team_Configuration;

   procedure Tactic_Change
     (Team_Id : in Team_Id_T;
      New_Formation_Id : in Positive;
      New_Pressing_Level : in Integer) is
   begin
      -- change the formation and the pressing value of the team
      Teams_Status(Team_Id).Set_Tactic(Formations_Array(New_Formation_Id),
                                       New_Pressing_Level);
      -- create and push the change tactic event and push it into the buffer
      Events_Buffer.Add(Change_Tactic_Event_T'(Team_Id        => Team_Id,
                                               Formation_Id   => New_Formation_Id,
                                               Pressing_Value => New_Pressing_Level));
      -- respond to the billboard with a "ok" response
      case Team_Id is
         when Local =>
            Notify_Response_Inquiry(Destination    => localManagerGUI,
                                    Reply_Type       => "tacticChange",
                                    Reply_Status   => True,
                                    Reply_Numeric_Detail => 0);
         when Visitor =>
            Notify_Response_Inquiry(Destination    => visitorManagerGUI,
                                    Reply_Type       => "tacticChange",
                                    Reply_Status   => True,
                                    Reply_Numeric_Detail => 0);
      end case;

      --@process request and response with a Remote Procedure Call
   end Tactic_Change;


   procedure Player_Substitution
     (Team_Id: in Team_Id_T;
      In_Player_Id : in integer;
      Out_Player_Id : in integer) is
   begin
      begin
         -- request the substitution to the referee
         Referee_Task.Substitution(
           Substitution_Request_T'(Team_Id               => Team_Id,
                                   Outgoing_Shirt_Number => Out_Player_Id,
                                   Incoming_Shirt_Number => In_Player_Id));
      exception
         when others =>
            null;
      end;
      -- resonse
      case Team_Id is
         when Local =>
            Notify_Response_Inquiry(Destination    => localManagerGUI,
                                    Reply_Type     => "playerSubstitutionAck",
                                    Reply_Status   => True,
                                    Reply_Numeric_Detail => 0);
         when Visitor =>
            Notify_Response_Inquiry(Destination    => visitorManagerGUI,
                                    Reply_Type     => "playerSubstitutionAck",
                                    Reply_Status   => True,
                                    Reply_Numeric_Detail => 0);
      end case;
   end Player_Substitution;
end Core_Listener;
