with Output_Types; use Output_Types;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Input_Types; use Input_Types;

package Distribution_Control is
   -- a record type to store the configuration of a player
   type Player_Configuration_T is record
      -- id of the player
      Player_Id : Unbounded_String := Null_Unbounded_String;
      -- index of the player role in the formation schema
      Player_Role_Index : Natural := 0;
      -- boolean to say if the player is first-string or substitute
      Is_First_String : Boolean := False;
   end record;

   -- an array type to store the players configuration for a team
   type Players_Configuration_T is array (Player_Number_T) of
     Player_Configuration_T;

   -- a record type to store the configuration of a team
   type Team_Configuration_T is record
      -- id of the team chosen
      Team_Id : Unbounded_String := Null_Unbounded_String;
      -- id of the formation chosen
      Formation_Id : Natural := 0;
      -- pressing value chosen
      Pressing_Balance_Value : Integer := 0;
      -- array of players chosen
      Players_Configuration : Players_Configuration_T;
   end record;

   -- a record type to store the configuration of the simulation
   type Simulation_Configuration_T is record
      -- the duration of the two match period (minutes)
      Period_Duration : Positive;
      -- the duration of the half time break (minutes)
      Break_Duration : Positive;
      -- the number of substitution allowed (per team)
      Substitutions_Allowed : Positive;
   end record;

   -- an array of two element to store teams configuration received by ManagerGUI
   type Teams_Configuration_T is array (Team_Id_T) of Team_Configuration_T;

   -- a protected object to control simulation configuration receiving
   protected Simulation_Configuration is
      -- a procedure used by main to initialize the protected object with the
      -- default configuration
      procedure Init(Default_Configuration : in Simulation_Configuration_T);
      -- an entry used by main procedure to wait for simulation start signal
      entry Wait_For_Simulation_Start;
      -- a procedure used by main procedure to stop configuration receiving;
      -- it is called after the delay expiration
      procedure Stop_Receiving;
      -- a procedure used by "distribution" to set the simulation configuration
      procedure Set(New_Configuration : in Simulation_Configuration_T;
                    Success : out Boolean);
      -- a function used by main procedure to get the simulation configuration
      function Get return Simulation_Configuration_T;
   private
      -- a boolean to say if the start signal has arrived (from SimulatorGUI)
      Start_Signal_Has_Arrived : Boolean := False;
      -- a boolean to say if the core can accept simulation configuration from
      -- SimulatorGUI, aka if the delay expired
      Can_Accept_Configuration : Boolean := True;
      -- the simulation configuration
      Configuration : Simulation_Configuration_T;
   end Simulation_Configuration;

   -- a protected object to control teams configuration receiving
   protected Teams_Configuration is
      -- a procedure used by "distribution" to set configuration for a team
      procedure Set(Team_Id		: in Team_Id_T;
                    Team_Configuration 	: in Team_Configuration_T;
                    Configuration_Code 	: out Configuration_Code_T);
      -- a procedure used by core to set the default configuration for a team
      procedure Set_Default(Team_Id		: in Team_Id_T;
                            Xml_Team_Id		: in Unbounded_String;
                            Formation_Id	: in Natural;
                            Pressing_Value	: in Integer);
      -- a function used by main procedure to get teams configuration
      function Get return Teams_Configuration_T;
      -- a procedure used by main procedure to stop receiving teams configuration
      -- from "distribution"
      procedure Stop_Receiving;
   private
      -- a boolean to say if core can accept team configuration from "distribution"
      Can_Accept_Configuration : Boolean := True;
      -- an array of record to store teams configuration
      Teams_Configuration : Teams_Configuration_T;
   end Teams_Configuration;

   -- a string which represent the simulation configuration XML tree
   Simulation_Xml_String : Unbounded_String;
   -- a string which represent the teams configuration XML tree
   Teams_Xml_String : Unbounded_String;
   -- simulation configuration
   Local_Simulation_Configuration : Simulation_Configuration_T;
end Distribution_Control;
