with Field; use Field;
with Ada.Numerics.Elementary_Functions;
use Ada.Numerics.Elementary_Functions;
with Status; use Status;
with Utils;

Package body Motion is

   function Get_Fixed_Motion(Position: Coordinates_3D_T) return Motion_3d_T is
   begin
      return Motion_3d_T'(T    => Time_Array'(
                          others => Status.Get_Corresponding_Quantum(Ada.Real_Time.Clock)),
                          X    => X_Array'(0 => Position.X, others => Position.X),
                          Y    => Y_Array'(0 => Position.Y, others => Position.Y),
                          Last => 0,
                          Z    => Z_Array'(0 => Position.Z, others => 0.0));
   end Get_Fixed_Motion;

   function Calculate_Movement(From: in Coordinates_T; 	-- initial position
                               To: in Coordinates_T; 	-- destination
                               Speed: Float)		-- speed in m/s
                               return Motion_T
   is
      Motion		: Motion_T;
      Pos_i		: Coordinates_T; -- Position corresponding to step i
      Dist		: Float;	 -- Distance done at step i
      Spent_Time	: Time_Span;	 -- The time needed to reach position i
   begin
      -- Position w/ index 0 represents the initial position
      Motion.X(0) := From.X;
      Motion.Y(0) := From.Y;
      Motion.T(0) := Status.Get_Corresponding_Quantum(Clock);
      Motion.Last := X_Range_T'Max(Abs(To.X - From.X),Abs(To.Y - From.Y));

      if Motion.Last = 0 then return Motion; end if;

      -- Calcolate the 2D motion step by step
      for i in 0..Motion.Last loop
         Motion.X(i) 	:= From.X + (To.X-From.X) * i / Motion.Last;
         Motion.Y(i) 	:= From.Y + (To.Y - From.Y) * i / Motion.Last;
         Pos_i 		:= Coordinates_T'(Motion.X(i),Motion.Y(i));
         Dist		:= Position.Distance(From, Pos_i);
         Spent_Time 	:= To_Time_Span(Duration(Dist / Speed));
         Motion.T(i)	:= Status.Get_Corresponding_Quantum(Motion.T(0) + Spent_Time);
      end loop;
      return Motion;
   end Calculate_Movement;

   function Get_Intersection(Player_Position: Coordinates_T;
                             Player_Speed: Float;	-- m/s
                             Player_Max_Height: Float;	-- m
                             Ball_Motion: Motion_3d_T)
                             return Coordinates_T is
      Dist: Float := Float'Last;
      Current : Coordinates_T;
   begin
      for i in 0..Ball_Motion.Last loop
         if Ball_Motion.Z(i) <= Player_Max_Height then
            Current := Coordinates_T'(Ball_Motion.X(i), Ball_Motion.Y(i));
            Dist := Distance(Player_Position, Current);
            if (Ada.Real_Time.Clock
                + Milliseconds(Integer(1000.0 * Dist / Player_Speed))
                < Ball_Motion.T(i)) then
               return Current;
            end if;
         end if;
      end loop;
      return Coordinates_T'(Ball_Motion.X(Ball_Motion.Last),
                            Ball_Motion.Y(Ball_Motion.Last));
   end;

   function Calculate_Ball_Motion (From: in Coordinates_3D_T;
                                   To: in Coordinates_3D_T;
                                   Initial_Speed: Float;
                                   Angle : Float := Ada.Numerics.Pi/6.0)
                                   return Motion_3d_T
   is
      Motion: Motion_3d_T;
      Pos_i		: Coordinates_3D_T; -- Position corresponding to step i
      Dist		: Float	:= 0.0;	    -- Distance done at step i
      Steps	: Positive;
      v, v_xy, v_z  : Float;
      Last_Bounce : Coordinates_3D_T := From;
      t_i  : Float;
      t_0  : Ada.Real_Time.Time;
   begin
      -- Position w/ index 0 represents the initial position
      Motion.X(0) := From.X;
      Motion.Y(0) := From.Y;
      Motion.Z(0) := From.Z;
      Motion.T(0) := Status.Get_Corresponding_Quantum(Clock);
      t_0 := Motion.T(0);

      -- after "Steps" steps the ball hits the ground
      -- for the 1st time (this could happen outside the pitch)
      Steps :=  X_Range_T'Min
        (X_Range_T'Last,
         X_Range_T'Max(Abs(To.X - From.X) + 1, Abs(To.Y - From.Y) + 1));
      Motion.Last := X_Range_T'Min(X_Range_T'Last, Steps*2);

      -- assuming alpha = 30
      v := Sqrt( (Distance(From, To)*9.81) / Sin(2.0 * Angle) );
      if v > Initial_Speed then
         v := Initial_Speed;
      end if;
      Utils.Logger.Add_Log("Ball speed: " & Integer(v)'Img & " m/s");
      v_xy := v * cos(angle);
      v_z  := v * sin(angle);
      -- Next_Bounce := 2.0 * v_xy * v_z / 9.81;

      -- while Dist < Next_Bounce
      for i in 1..X_Range_T'Last
      loop	-- could go out-of-range...
         Motion.X(i) := From.X + (To.X-From.X) * i / Steps;
         Motion.Y(i) := From.Y + (To.Y - From.Y) * i / Steps;
         Pos_i 	:= Coordinates_3D_T'(Motion.X(i), Motion.Y(i), Motion.Z(i));
         Dist	:= Position.Distance(Last_Bounce, Pos_i);

         -- t_i = s_xy(t_i) / (v(t_0)*cos(alpha))
         t_i := Dist / v_xy;
         Motion.T(i) := Get_Corresponding_Quantum
           (t_0 + Milliseconds(Integer(t_i * 1000.0)));
         Motion.Z(i)	:= Float'Max(0.0, v_z * t_i - 1.0/2.0 * 9.81 * t_i**2);

         if Motion.Z(i) = 0.0 then
            v := v * 0.8; -- speed slows down by 20% at each bounce
            v_xy := v * cos(angle);
            v_z  := v * sin(angle);
            t_0  := Motion.T(i);
            Last_Bounce := (Motion.X(i), Motion.Y(i), Motion.Z(i));
         end if;

         if (Contains(Pitch, Coordinates_T'(Motion.X(i-1),Motion.Y(i-1)))
             and not Contains(Pitch, Coordinates_T'(Motion.X(i),Motion.Y(i))))
           or Motion.T(i) > Motion.T(i-1) + Seconds(2) -- stops when speed < 0.5 m/s
         then
            Motion.Last := i;	-- used to set Last and to exit the loop
            exit;
         end if;
      end loop;
      return Motion;
   end Calculate_Ball_Motion;

   function To_Flat_Motion_3d(Motion : Abstract_Motion_T'Class)
                              return Motion_3d_T is
   begin
      return Motion_3d_T'(Abstract_Motion_T(Motion) with Z_Array'(others => 0.0));
   end;

   function Last(Motion: Abstract_Motion_T'Class) return X_Range_T is
   begin
      return Motion.Last;
   end Last;

   function Element_At (Motion:Motion_T; Index: X_Range_T)
                        return Coordinates_T is
   begin
      return Coordinates_T'(Motion.X(Index), Motion.Y(Index));
   end Element_At;

   function Time_At (Motion:Abstract_Motion_T'Class; Index: X_Range_T)
                     return Time is
   begin
      return Motion.T(Index);
   end;

   function Element_At (Motion	: Motion_3d_T; Index : X_Range_T)
                        return Coordinates_3d_T is
   begin
      return Coordinates_3d_T'(Motion.X(Index),Motion.Y(Index),Motion.Z(Index));
   end Element_At;

   function Get_Current_Position (Motion: Motion_3d_T)
                                  return Coordinates_3D_T is
      Current_Time : constant Time := Get_Corresponding_Quantum(Clock);
      L: X_Range_T := 0;		-- left bound
      R: X_Range_T := Motion.Last;	-- right bound
      Index : X_Range_T := (R + L) / 2;	-- (currently) selected index
   begin
      while L < R - 1 loop
         Index := (R + L) / 2;
         -- If the element in Index comes after Current_Time
         if Motion.Time_At(Index) > Current_Time then
            -- choose the first half of the current selection
            R := Index; -- i.e. the right bound is set to Index
         elsif	-- otherwise, if it comes before, choose the second half
           Motion.Time_At(Index) <= Current_Time then
            L:= Index;
         end if;
      end loop;
      if Motion.Time_At(R) <= Current_Time
      then return Motion.Element_At(R);
      else return Motion.Element_At(L);
      end if;
   end;
end Motion;
