with Position; use Position;
with Ada.Real_Time; use Ada.Real_Time;
with Ada.Numerics;
with Output_Types; use Output_Types;

Package Motion is
   ---------------------------------------------------------------
   --				MOTION				--
   ---------------------------------------------------------------
   -- abstract motion
   type Abstract_Motion_T is abstract tagged private;

   -- a private operation to get the last element of a 2d motion
   function Last (Motion : in Abstract_Motion_T'Class) return X_Range_T;

   -- a private operation to get the time component of an element from a 2d motion
   function Time_At (Motion	: Abstract_Motion_T'Class;
                     Index	: X_Range_T) return Time;

   -- a type to represent a 2d motion
   type Motion_T is new Abstract_Motion_T with private;

   -- a function to get a position in a discrete motion
   function Element_At (Motion	: Motion_T;
                        Index 	: X_Range_T) return Coordinates_T;

   -- a function to calculate the motion needed to make a movement
   function Calculate_Movement
     (From: in Coordinates_T; To: in Coordinates_T; Speed: Float)
      return Motion_T;

   ---------------------------------------------------------------
   --				MOTION 3D			--
   ---------------------------------------------------------------
   -- a subtype of Motion_T to represent a 3d motion
   type Motion_3d_T is new Abstract_Motion_T with private;

   -- a private operation to get an element from a 3d motion (by its index)
   function Element_At (Motion	: Motion_3d_T;
                        Index 	: X_Range_T) return Coordinates_3D_T;

   -- the 'motion' of an object that is not moving
   function Get_Fixed_Motion (Position: Coordinates_3D_T) return Motion_3d_T;

   -- given a motion, return the current (3d) position
   -- this function is not required for a 2-dimensional motion, because players
   -- do not need to look into their calculated motion to get their position
   function Get_Current_Position(Motion: Motion_3d_T) return Coordinates_3D_T;

   -- a function to get the first position in which the player could be able
   -- to reach and catch a moving ball. It might fail if there are obstacles
   -- in his path
   function Get_Intersection(Player_Position: Coordinates_T;
                             Player_Speed: Float;	-- m/s
                             Player_Max_Height: Float;	-- m
                             Ball_Motion: Motion_3d_T)
                             return Coordinates_T;

   -- a function to calculate the ball motion
   function Calculate_Ball_Motion(From: in Coordinates_3D_T;
                                  To: in Coordinates_3D_T;
                                  Initial_Speed: Float;
                                  Angle : Float := Ada.Numerics.Pi/6.0)
                                  return Motion_3d_T;

   function To_Flat_Motion_3d(Motion : Abstract_Motion_T'Class) return Motion_3d_T;
private
   type X_Array is array (X_Range_T) of X_Range_T;
   type Y_Array is array (X_Range_T) of Y_Range_T;
   type Z_Array is array (X_Range_T) of Float;
   type Time_Array is array(X_Range_T) of Ada.Real_Time.Time;

   type Abstract_Motion_T is abstract tagged record
      T     : Time_Array; -- t(i) time needed to reach the next position
      X     : X_Array;
      Y     : Y_Array;
      Last  : X_Range_T;  -- index of the last position in the whole movement
      -- subsequent array elements are ignored (0 refers to the first pos.)
      -- After the last time value, the corresponding item is supposed to mantain
      -- its last position (permanently)
   end record;

   type Motion_T is new Abstract_Motion_T with null record;

   type Motion_3d_T is new Abstract_Motion_T with record
      Z : Z_Array := (others => 0.0);
   end record;
end Motion;
