with Field; use Field;
with Tactic; use Tactic;
with Motion; use Motion;
with Game_Control; use Game_Control;
with Ada.Real_Time; use Ada.Real_Time;
with Status;
with Status.Manager; Use Status.Manager;
with Ball; use Ball;
with Utils;
use Utils;
with Field_Control;
with Television;
with Ada.Numerics.Float_Random; use Ada.Numerics.Float_Random;
with Events; use Events;
with Events.Player_Substitution; use Events.Player_Substitution;
with Events.Throw;
with Position; use Position;
with Player; use Player;

Package body Player.Logic is
   Throw_Time	: constant Time_Span := Milliseconds(400);
   Catch_Time	: constant Time_Span := Milliseconds(400);
   Tackle_Time	: constant Time_Span := Milliseconds(400);
   Idle_Time	: constant Time_Span := Milliseconds(100);
   Penalty_Time	: constant Time_Span := Milliseconds(1500);
   -- penalty time is applied to a player who has lost the ball possession
   -- after being tackled

   -- matrix used to map a player's neighbourhood. For each position,
   -- True means that a position is occupied by another player, False that it's free
   type Boolean_Matrix is array(Integer range <>, Integer range <>) of Boolean;

   -- a Matrix used to store all the position of other players when they are read
   type Coordinates_Matrix is array(Team_Id_T, Player_Index_T) of Coordinates_T;

   -- the closest players of each team
   type Closest_Players_T is array(Team_Id_T) of Player_Number_T;

   -- all the information about others (and about the ball) read by the player
   -- to make his choice
   type Players_Info_T is record
      Positions		   : Coordinates_Matrix;
      Neighborhood	   : Boolean_Matrix(-1..1,-1..1);
      Ball_Closest_Players : Closest_Players_T;
      Pos_Closest_Players  : Closest_Players_T;
   end record;

   type Interruption_Handling_Phase_T is (Not_Handling, Player_Condition, Game_Interruption);

   -- read information about ball and other players' position
   procedure Read_Players_Info(Players_Info	: in out Players_Info_T;
                               My_Position	: in Coordinates_T;
                               Ball_Position	: in Coordinates_3D_T)
   is
      -- the position currently read
      Pos: Coordinates_T;
      -- the distance from *this* (executing) player to the closest one
      Pos_Closest_Distance	: Float;
      -- the distance of the ball's closest player to the ball itself
      Ball_Closest_Distance	: Float;
      -- the distance of the player currently being read to the ball
      Current_Distance		: Float;
   begin
      -- Improvement: Calculate Ball_Closest_Teammate e My_Closest_Opponent
      Players_Info.Neighborhood := (others => (others => True));
      for i in Team_Id_T loop
         -- Reset the following values because they are used for both teams
         Pos_Closest_Distance := Float'Last;
         Ball_Closest_Distance := Float'Last;
         for j in Player_Index_T loop
            -- Read the position of Player_ij (access a PO)
            Pos := Status.Player_Mapper(i,j).Get_Position;
            -- Store his position ion Player_Info for future use
            Set_Position(Players_Info.Positions(i, j),Pos.X, Pos.Y);
            -- compare distance against My_Position
            Current_Distance := Distance(My_Position, Pos);
            if Current_Distance < Pos_Closest_Distance
              and then Current_Distance > 1.0
            -- otherwise the player might find himself as the closest one
            -- The lazy evaluation above often avoids the second check
            then
               -- Found a player for team i more closed to My_Position than the previous
               Pos_Closest_Distance := Current_Distance;
               Players_Info.Pos_Closest_Players(i) :=
                 Status.Player_Mapper(i,j).Get_Player_Number;
            end if;
            -- calculate distance to the ball
            Current_Distance := Distance(Pos, Coordinates_T(Ball_Position));
            if Current_Distance < Ball_Closest_Distance then
               -- Found a player for team i more closed to the ball than the previous
               Ball_Closest_Distance := Current_Distance;
               Players_Info.Ball_Closest_Players(i) :=
                 Status.Player_Mapper(i,j).Get_Player_Number;
            end if;
            -- calculate the available positions in the neighborhood
            -- PRE: for all a in [Team_Id_T'First..i-1] and b in
            -- [Player_Index_T'First..j-1], let "Pos_ab" be the position
            -- of the player b of the team a and "My" the position of
            -- the calling player. Then if Pos_ab is nearby (My +/- 1 cell)
            --  Neighborhood(Pos.X,Pos.Y) is False.
            if Pos.X in My_Position.X - 1 .. My_Position.X + 1
              and Pos.Y in My_Position.Y - 1 .. My_Position.Y + 1
            then
               Players_Info.Neighborhood
                 (My_Position.X - Pos.X, My_Position.Y - Pos.Y) := False;
               -- now PRE is also valid for Pos_ij
            end if;
         end loop; -- j in Player_Index_T
      end loop; -- i in Team_Id_T
   end Read_Players_Info;

   function Closest_To_Position(Player_Info: Players_Info_T;
                                Team: Team_Id_T;
                                Position: Coordinates_T) return Player_Number_T
   is
      Min_Distance : Float := Float'Last;
      Curr_Distance : Float;
      Closest_Player : Player_Number_T;
   begin
      for i in Player_Index_T loop
         Curr_Distance := Distance(Player_Info.Positions(Team, i),Position);
         if Curr_Distance < Min_Distance then
            Min_Distance := Curr_Distance;
            Closest_Player := Status.Player_Mapper(Team, i).Get_Player_Number;
         end if;
      end loop;
      return Closest_Player;
   end;

   -- *** TASK BODY ***
   task body Player_Task_T is
      -- PLAYER IDENTITY
      My_Id: Identifier_T := (Team_Id => Tid, Player_Number => Pid);

      -- STATUS
      -- Player status
      My_Position	: Coordinates_T;	-- to be read from Mapper
      Player_State	: Player_State_T := Active;
      Player_Characteristics   : Player_Characteristics_T;
      Match_Finished	: Boolean 	:= False; -- exits when finished

      -- Team status
      Formation	: Formation_T;
      Side	: Side_T;

      -- Others
      Ball_Status	: Ball_Status_T;
      Ball_Status_Changed : Boolean := True;
      Former_Game_State	: Game_Control.State_Record_T;
      Game_State	: Game_Control.State_Record_T;
      Players_Info	: Players_Info_T;

      -- LOCAL
      Movement		: Motion_T;
      Current_Step	: Natural := 0;
      Next_Execution	: Ada.Real_Time.Time;
      Init_Time		: Ada.Real_Time.Time;
      Has_Field_Lock	: Boolean := False;
      Ball_Possession	: Boolean := False;

      -- *** Action handling
      -- Player transits along three machine states while handling interruptions
      Interruption_Handling_Phase : Interruption_Handling_Phase_T := Not_Handling;
      Moving_Somewhere	: Boolean := False; -- last destination not yet reached

      -- True if player should "leave" the barrier by itself
      Should_Release	: Boolean := True;

      -- Adapter procedure to execute the actions required to change the
      -- player' position (accessing the status manager).
      procedure Move_To (New_Position: Coordinates_T) is
      begin
         Status.Manager.Manager.Move(Team_Id => Tid,
                                     Task_Id => Pid,
                                     New_Position      => New_Position);
         My_Position := New_Position;
         Should_Release := False;
      end Move_To;

      -- avoid multiple invocation eventually due to the task complexity
      procedure Safely_Leave_Field_Control is
      begin
         if Should_Release then
            Field_Control.Barrier.Leave(My_Position);
         end if;
         Has_Field_Lock := False;
      end Safely_Leave_Field_Control;

      -- There are many points in which a player could need to take control
      -- of the positions around him. This procedure avoid multiple requests
      -- to Barrier.Play.
      procedure Get_Field_Control is
      begin
         Init_Time := Status.Get_Corresponding_Quantum(Clock);
         if not Has_Field_Lock then
            select
               Field_Control.Barrier.Play(My_Position);
            or delay 1.0;
               if My_Id.Team_Id = Local and Pid = 2 then
                  Logger.Add_Log("PROCEED W/OUT FIELD CONTROL - Player "
                                 & My_Id.Team_Id'Img & My_Id.Player_Number'Img);
               end if;
            end select;
         end if;
         Next_Execution := Init_Time + Idle_Time;
         My_Position := Status.Player_Mapper(Tid,Pid).Get_Position;
         Has_Field_Lock := True;
         Should_Release := True;
      end Get_Field_Control;

      -- Set up the initial player configuration
      procedure Initialize is
      begin
         My_Id.Team_Id := Tid;
         My_Id.Player_Number := Status.Player_Mapper(Tid, Pid).Get_Player_Number;
         Side := Status.Teams_Status(My_Id.Team_Id).Get_Side;
         Game_State := Game_Control.Game_State.Get;
         My_Position := Status.Player_Mapper(Tid, Pid).Get_Position;
         Player_Characteristics := Status.Players_Characteristics(My_Id.Team_Id,
                                                                  My_Id.Player_Number);
      end;

      -- PRE: Formation, Side, Role, Game_State and Ball_Position are up to date
      procedure Calculate_Movement is
         Pressing : constant Integer := Status.Teams_Status(Tid).Get_Pressing;
      begin
         Formation := Status.Teams_Status(Tid).Get_Formation;
         -- Check whether the player should leave the field or not
         if (Game_State.State /= Play)
           and (Player_State in Injured..Sent_Off) then
            Movement := Motion.Calculate_Movement
              (From => My_Position,
               To   => Tactic.Outside_Position
                 (Side 	     => Side,
                  Role_Index => Pid),
               Speed => Player_Characteristics.Speed);
         else -- the position depends on the game state and ball position
            case Game_State.State is
            when Play =>
               declare
                  Intersection : Coordinates_T;
                  Closest_Player: Player_Number_T;
                  Current_Ball_Position: Coordinates_3D_T;
                  Predicted_Motion: Motion_T;
               begin
                  if Ball_Status.Free = False
                    and Ball_Status.Possessor.Team_Id /= My_Id.Team_Id
                  then

                     Current_Ball_Position :=
                       Motion.Get_Current_Position(Ball_Status.Motion);
                     -- Predict the ball movement (it moves with the opponent)
                     if Side = Left then
                        Predicted_Motion := Motion.Calculate_Movement
                          (From  => Coordinates_T(Current_Ball_Position),
                           To    => Coordinates_T'(Pitch.Left, Current_Ball_Position.Y),
                           Speed => Player_Characteristics.Speed);
                     else
                        Predicted_Motion := Motion.Calculate_Movement
                          (From  => Coordinates_T(Current_Ball_Position),
                           To    => Coordinates_T'(Pitch.Right, Current_Ball_Position.Y),
                           Speed => Player_Characteristics.Speed);
                     end if;
                     -- Watch where are the other players
                     Read_Players_Info
                       (Players_Info  => Players_Info,
                        My_Position   => My_Position,
                        Ball_Position => (Motion.Element_At
                                          (Motion => Predicted_Motion,
                                           Index  => Motion.Last(Predicted_Motion))
                                          with 0.0));
                     -- Calculate where the player could intercept the ball
                     Intersection:= Motion.Get_Intersection
                       (Player_Position   => My_Position,
                        Player_Speed      => Player_Characteristics.Speed,
                        Player_Max_Height => 20.0, -- TODO
                        -- Player_Characteristics.Height + Player_Characteristics.Jump,
                        Ball_Motion       => To_Flat_Motion_3d(Predicted_Motion));

                  else	-- nobody has ball possession
                     -- Watch where are the other players
                     Read_Players_Info
                       (Players_Info  => Players_Info,
                        My_Position   => My_Position,
                        Ball_Position => Motion.Element_At
                          (Motion => Ball_Status.Motion,
                           Index  => Motion.Last(Ball_Status.Motion)));
                     -- Look for a position in which the ball could be intercepted
                     Intersection:= Motion.Get_Intersection
                       (Player_Position   => My_Position,
                        Player_Speed      => Player_Characteristics.Speed,
                        Player_Max_Height => Player_Characteristics.Height +
                          Player_Characteristics.Jump_Skill,
                        Ball_Motion       => Ball_Status.Motion);
                  end if;

                  -- Check who in the player's team is closer to the ball
                  Closest_Player :=
                    Closest_To_Position(Player_Info => Players_Info,
                                        Team        => My_Id.Team_Id,
                                        Position    => Intersection);

                  -- If the player doesn't have the ball possession
                  if not Ball_Possession then
                     if Ball_Status.Free = False
                       and Ball_Status.Possessor.Team_Id /= My_Id.Team_Id
                       and Status.Players_Status
                     -- which is held by the opponents' goalkeeper
                       (Ball_Status.Possessor.Team_Id,
                        Ball_Status.Possessor.Player_Number).Get_Role = 1
                     then -- move back to the "standard position"
                        Movement := Motion.Calculate_Movement
                          (From  => My_Position,
                           To    => Tactic.Get_Standard_Position
                             (Formation, Side, Role_Index => Pid),
                           Speed => Player_Characteristics.Speed);
                        -- otherwise, if the player doesn't have the ball possession
                        -- but he's the closest to the ball
                     elsif (Closest_Player = My_Id.Player_Number) then
                        -- Move toward the ball
                        Movement := Motion.Calculate_Movement
                          (From => My_Position, To => Intersection,
                           Speed => Player_Characteristics.Speed);
                     else -- player doesn't have the ball possession
                        -- and is not the closest to the ball
                        -- => Move toward the "reference position"
                        Movement :=
                          Motion.Calculate_Movement
                            (From => My_Position,
                             To   => Tactic.Reference_Position
                               (Formation 	=> Formation,
                                Side 	=> Side,
                                Role_Index 	=> Pid,
                                State 	=> Game_State.State,
                                Ball_Position => Motion.Element_At
                                  (Ball_Status.Motion, Ball_Status.Motion.Last),
                                Pressing => Pressing),
                             Speed => Player_Characteristics.Speed);
                     end if;
                  else -- if the player Has_The_Ball, run toward the goal
                     if Side = Left then
                        Movement := Motion.Calculate_Movement
                          (From => My_Position,
                           To => Coordinates_T'(Pitch.Right, My_Position.Y),
                           Speed => Player_Characteristics.Speed);
                     else
                        Movement := Motion.Calculate_Movement
                          (From => My_Position,
                           To => Coordinates_T'(Pitch.Left, My_Position.Y),
                           Speed => Player_Characteristics.Speed);
                     end if;
                  end if;

               end; -- of declare and also END PLAY

               -- IN CASE OF:
               -- Kick_Off, Throw_In, Goal_Kick, Corner_Kick, Free_Kick, Penalty_Kick
            when Kick_Off..Penalty_Kick =>
               if Game_State.Offence_Team = Tid
                 and Status.Teams_Status(Tid).Get_Starting_Player = Pid
               then
                  Movement := Motion.Calculate_Movement
                    (From => My_Position,
                     To   => Coordinates_T(Game_State.Position),
                     Speed => Player_Characteristics.Speed);
               else
                  Movement :=
                    Motion.Calculate_Movement
                      (From => My_Position,
                       To   => Tactic.Reference_Position
                         (Formation	=> Formation,
                          Side		=> Side,
                          Role_Index 	=> Pid,
                          State	 	=> Game_State.State,
                          Ball_Position 	=> Game_State.Position,
                          Pressing => Pressing),
                       Speed => Player_Characteristics.Speed);
               end if; -- Calculate the right position to restart the game
            when Half_Time_Break..End_Match =>
               Side := Status.Teams_Status(Tid).Get_Side;
               Movement :=
                 Motion.Calculate_Movement
                   (From => My_Position,
                    To   => Tactic.Reference_Position
                      (Formation	=> Formation,
                       Side		=> Side,
                       Role_Index 	=> Pid,
                       State	 	=> Game_State.State,
                       Ball_Position => Game_State.Position,
                       Pressing => Pressing),
                    Speed => Player_Characteristics.Speed);
            when Not_Started =>	-- should never happen
               -- polling avoided by the initialization protocol
               Next_Execution:=
                 Status.Get_Corresponding_Quantum(Init_Time + Idle_Time);
            end case;
         end if; -- (Player_State in Injured..Sent_Off) - else branch
         if Movement.Last = 0 then
            -- The player is already in his correct position and should
            -- stay there
            Moving_Somewhere := False;
            -- Let the player be idle for a while...
            -- The delay is required because no player is allowed
            -- to execute twice in the same quantum)
            Next_Execution :=
              Status.Get_Corresponding_Quantum(Init_Time + Idle_Time);
         else
            -- prepare to wait the time needed to make the first step
            Moving_Somewhere := True;
            Current_Step := 1;
            Next_Execution :=
              Status.Get_Corresponding_Quantum(Motion.Time_At(Movement, Current_Step));
         end if;
      end;

      procedure Make_Next_Step is
         -- Position is the current position of the player, before this step
         -- New_Position is the new one taken by the player through this step
         New_Position : Coordinates_T;
         -- Next_Position is the one to be taken through the next step
         Next_Position: Coordinates_T;
         -- True if it exist a position, different than New_Position, from
         -- which it is possible to reach Next_Position in one single step
         Has_Alternative : Boolean := True;
         -- X offset between Position and Next Position
         Delta_X : Natural := 0;
         -- Y offset between Position and Next Position
         Delta_Y : Natural := 0;
         -- Region of the field made with positions closed max 1 step from
         -- both Position and Next_Position
         Common_Neighborhood : Field_Area_T;
      begin
         New_Position := Motion.Element_At(Movement, Current_Step);
         if (Current_Step /= Movement.Last) then
            Next_Position := Motion.Element_At(Movement, Current_Step + 1);
         else
            Next_Position := New_Position; -- NOTE: No need to keep moving !!
         end if;
         -- Delta_X (Y) is the distance along the X (Y) axis between the actual
         -- position (before the current step) and the position after the
         -- next step.
         Delta_X := abs(Next_Position.X - My_Position.X);
         Delta_Y := abs(Next_Position.Y - My_Position.Y);
         if (Delta_X = 2 and Delta_Y = 2) then
            Has_Alternative := False;
            -- When a player tries movements like: (1,1) -> (2,2) -> (3,3)
            -- and (2,2) is busy, there is no way to avoid an obstacle
            -- mantaining the same number of steps
         end if;
         -- Calculate the set of positions closed 1 step from both the current
         -- Position and the Next_Position. That set of positions is enclosed
         -- into the Common_Neighborhood area.
         -- This calculation aims to avoid an obstacle while preserving
         -- the calculated movement of the player: using Common_Neighborhood
         -- the player can make a 1-step exception to his trajectory without
         -- compromising the whole movement.

         if Delta_X = 0 then
            -- the player is supposed to make two steps along the Y axis,
            -- so he could avoid an obstacle by moving one step left or right
            Common_Neighborhood.Left :=
              X_Range_T'Max(X_Range_T'First, My_Position.X - 1);
            Common_Neighborhood.Right :=
              X_Range_T'Min(X_Range_T'Last, My_Position.X + 1);
         elsif Delta_X = 1 then
            -- one of the two steps is also along the X axis: it could be done
            -- immediately or during the next step
            Common_Neighborhood.Left := X_Range_T'Min(My_Position.X, Next_Position.X);
            Common_Neighborhood.Right := X_Range_T'Max(My_Position.X, Next_Position.X);
         else -- Delta_X = 2
            -- both the two steps are also along the X axis, so the movement
            -- along that axis can not be deferred
            Common_Neighborhood.Left := New_Position.X;
            Common_Neighborhood.Right := New_Position.X;
         end if;
         -- this block works as the one before...
         if Delta_Y = 0 then
            -- W/out the check on Y_Range_T the calculated area would expand
            -- out of the boundaries, raising an exception/killing the task.
            Common_Neighborhood.Top
              := Y_Range_T'Max(Y_Range_T'First, My_Position.Y - 1);
            Common_Neighborhood.Bottom
              := Y_Range_T'Min(Y_Range_T'Last, My_Position.Y + 1);
         elsif Delta_Y = 1 then
            Common_Neighborhood.Top := Y_Range_T'Min(My_Position.Y, Next_Position.Y);
            Common_Neighborhood.Bottom := Y_Range_T'Max(My_Position.Y, Next_Position.Y);
         else -- Delta_Y = 2
            Common_Neighborhood.Top := New_Position.Y;
            Common_Neighborhood.Bottom := New_Position.Y;
         end if;
         -- if Has_Alternative is False, accept coordinates which invalidate
         -- the whole movement (a new one will be then calculated)
         -- The new accepted coordinates are the two which share the X or Y
         -- coordinate with the current Position, and the one which shares both.
         -- NOTE: the last one must be chosen *only* if the other two are busy
         -- (because it means that the player won't advance)
         if not Has_Alternative then
            if Common_Neighborhood.Left > My_Position.X then
               Common_Neighborhood.Left := My_Position.X;
            elsif Common_Neighborhood.Right < My_Position.X then
               Common_Neighborhood.Right := My_Position.X;
            end if;
            if Common_Neighborhood.Top > My_Position.Y then
               Common_Neighborhood.Top := My_Position.Y;
            elsif Common_Neighborhood.Bottom < My_Position.Y then
               Common_Neighborhood.Bottom := My_Position.Y;
            end if;
         end if;

         -- Instantiate a boolean matrix to check for free positions inside
         -- the Common Neighborhood area.
         declare
            Available:
            Boolean_Matrix(Common_Neighborhood.Left..Common_Neighborhood.Right,
                           Common_Neighborhood.Top..Common_Neighborhood.Bottom);
            Found: Boolean := False;
         begin
            -- read the position of each player and select those which could
            -- interfere with the current step

            -- *** MUST HAVE FIELD CONTROL WHILE READING ***
            Read_Players_Info(Players_Info  => Players_Info,
                              My_Position   => My_Position,
                              Ball_Position => Motion.Element_At
                                (Ball_Status.Motion, Ball_Status.Motion.Last));

            for i in Common_Neighborhood.Left..Common_Neighborhood.Right loop
               for j in Common_Neighborhood.Top..Common_Neighborhood.Bottom loop
                  Available(i,j) :=
                    Players_Info.Neighborhood (My_Position.X - i, My_Position.Y - j);
               end loop;
            end loop;

            Found := False;
            if Available(New_Position.X, New_Position.Y) then
               -- the position to be taken during this step is the one
               -- precalculated during the movement.
               Found := True;
            else -- look for an alternative
               for i in Common_Neighborhood.Left..Common_Neighborhood.Right loop
                  for j in Common_Neighborhood.Top..Common_Neighborhood.Bottom
                  loop
                     if Available(i, j) then
                        New_Position := Coordinates_T'(i, j);
                        if My_Position /= New_Position then
                           -- When Has_Alternative is False, Common_Neighborhood
                           -- also contains Position; when chosen, the loops are
                           -- not interrupted because a different one could be
                           -- found later and the current Position should be
                           -- kept only as the last chance.
                           Found := True;
                           if not Has_Alternative then
                              Movement:=
                                Calculate_Movement
                                  (From  => New_Position,
                                   To    => Movement.Element_At(Movement.Last),
                                   Speed => Player_Characteristics.Speed);
                              Current_Step := 0;
                           end if;
                        end if; -- Position /= New_Position
                     end if; -- Available(i, j)
                     exit when Found; -- no need to find out any other free position
                  end loop;
                  exit when Found; -- no need to find out any other free position
               end loop;
            end if;
            if not Found then
               New_Position := My_Position;
            end if;
         end;
         if My_Position = New_Position then
            Movement:=
              Calculate_Movement
                (From  => My_Position,
                 To    => Movement.Element_At(Movement.Last),
                 Speed => Player_Characteristics.Speed);
            Current_Step := 0;
         else
            Move_To(New_Position);
         end if;
         if Current_Step = Movement.Last then
            Moving_Somewhere := False;
            -- In this case the player is allowed to execute twice in one quantum
         else
            Current_Step := Current_Step + 1;
            Next_Execution :=
              Status.Get_Corresponding_Quantum(Motion.Time_At(Movement, Current_Step));
         end if;
      end Make_Next_Step;

      -- Look for the closest teammate (to pass him the ball), possibly
      -- at least two steps forward and return a position more advanced
      -- in which he could play the ball
      procedure Look_For_Teammate
        (Teammate_Position : in out Coordinates_T;
         Game_State : in Game_Control.State_T := Play)
      is
         Min_Distance : Float := Float'Last;
         Curr_Distance : Float;
         Teammate_Area: Field_Area_T := Pitch;
      begin
         if Game_State = Play then
            if Side = Left then
               Teammate_Area.Left := My_Position.X - 2;
            else
               Teammate_Area.Right := My_Position.X + 2;
            end if;
         end if;

         for i in Player_Index_T loop
            Curr_Distance
              := Distance(Players_Info.Positions(My_Id.Team_Id, i),
                          My_Position);
            if i /= Pid and Curr_distance < Min_Distance and Curr_Distance > 2.0
              and Field.Contains
                (Area => (Teammate_Area.Left, Teammate_Area.Right,
                          Pitch.Top, Pitch.Bottom),
                 Position => Players_Info.Positions(My_Id.Team_Id, i))
            then
               Min_Distance := Curr_Distance;
               Teammate_Position :=
                 Players_Info.Positions(My_Id.Team_Id, i);
               if Side = Left then
                  Teammate_Position.X := Integer'Min
                    (Pitch.Right,
                     Teammate_Position.X + Integer(Curr_Distance / 2.5));
               else
                  Teammate_Position.X := Integer'Max
                    (Pitch.Left,
                     Teammate_Position.X - Integer(Curr_Distance / 2.5));
               end if;
            end if;
         end loop;
      end Look_For_Teammate;

      procedure Catch is
         G: Ada.Numerics.Float_Random.Generator;
      begin
         Reset(G);
         -- STOP/CATCH THE BALL (with some probability...)
         Ball_Possession := Random(G) < Player_Characteristics.Catching_Accuracy;
         Status.Manager.Manager.Catch(Tid, Pid, Ball_Possession);

         Calculate_Movement;

         if Ball_Possession then
            Next_Execution:= Status.Get_Corresponding_Quantum(Init_Time + Status.Quantum);
            -- Needed afterwards to check if the player has been tackled
            Ball_Status := -- motion is not actually relevant
              (Motion.Get_Fixed_Motion((My_Position with 0.0)), My_Id, False);
         else
            Next_Execution:= Status.Get_Corresponding_Quantum(Init_Time + Catch_Time);
         end if;
         Should_Release := False;
      end Catch;

      procedure Tackle is
         G: Ada.Numerics.Float_Random.Generator;
         Dribble_Value : constant Float :=
           Status.Players_Characteristics(Ball_Status.Possessor.Team_Id,
                                          Ball_Status.Possessor.Player_Number).Dribble_Accuracy;
         Is_Flawed : Boolean;
         Tackle_Damage : Float := 0.0;
      begin
         Reset(G);
         -- Ball_Possession is set to true if the tackle succeeds
         Ball_Possession :=
           Random(G) < (Player_Characteristics.Tackle_Skill + (1.0 - Dribble_Value)) / 2.0;
         Is_Flawed := Random(G) < Player_Characteristics.Aggression;
         if Is_Flawed then
            Tackle_Damage := Random(G) * (100.0 - Status.Foul_Limit) + Status.Foul_Limit;
         else
            Tackle_Damage := Random(G) * (Status.Foul_Limit);
         end if;

         Status.Manager.Manager.Tackle
           (Team_Id          => Tid,
            Player_Task_Id   => Pid,
            Offended_Shirt   => Ball_Status.Possessor.Player_Number,
            Tackle_Damage    => Tackle_Damage,
            Success          => Ball_Possession);
         if Ball_Possession then
            -- Needed afterward to check if the player has been tackled
            Ball_Status := -- motion is not actually relevant
              (Motion.Get_Fixed_Motion((My_Position with 0.0)), My_Id, False);
         end if;

         Moving_Somewhere := False;
         Calculate_Movement;
         Next_Execution:=
           Status.Get_Corresponding_Quantum(Init_Time + Tackle_Time);
         Should_Release := False;
      end Tackle;

      procedure Throw(Destination : in Coordinates_3D_T;
                      Angle : Float := Ada.Numerics.Pi/6.0) is
         Initial_Speed: Float;
         Initial_Ball_Height: constant Float :=
           Motion.Get_Current_Position(Ball_Status.Motion).Z;
         Ball_Motion : Motion_3d_T;
         G: Ada.Numerics.Float_Random.Generator;
         Accuracy : Float range 0.0..1.0;
         Local_Destination: Coordinates_3D_T := Destination;
         Category : Events.Throw.Category_T;
         Mode : Events.Throw.Mode_T;
      begin
         Reset(G);
         -- get the throw category
         if Destination.X < Pitch.Left or Destination.X > Pitch.Right then
            -- destination is out of goal line so the throw is a shot
            Category := Events.Throw.Shooting;
         else
            Category := Events.Throw.Passing;
         end if;
         -- get the throw mode
         if Field.Contains(Pitch,My_Position) then
            -- player is inside the game field, check the height
            if Initial_Ball_Height >= Player_Characteristics.Height then
               -- heading
               Mode := Events.Throw.Heading;
            else
               -- kick
               Mode := Events.Throw.Kick;
            end if;
         else
            -- player is outside the game field, so the throw is a throw-in
            Mode := Events.Throw.Throw_In;
         end if;

         if Initial_Ball_Height > Player_Characteristics.Height
         then
            Initial_Speed := Player_Characteristics.Heading_Power;
            Accuracy := Player_Characteristics.Heading_Accuracy;
         elsif not Field.Contains(Pitch, Destination) then
            Initial_Speed := Player_Characteristics.Shooting_Power;
            Accuracy := Player_Characteristics.Shooting_Accuracy;
         else
            Initial_Speed := Player_Characteristics.Passing_Power;
            Accuracy := Player_Characteristics.Passing_Accuracy;
         end if;
         Local_Destination.X := Local_Destination.X
           + Integer(Float(My_Position.X - Local_Destination.X)
                     * (Random(G) * 2.0 - 1.0) * (1.0 - Accuracy));
         Local_Destination.Y := Local_Destination.Y
           + Integer(Float(My_Position.Y - Local_Destination.Y)
                     * (Random(G) * 2.0 - 1.0) * (1.0 - Accuracy));


         Ball_Motion := Motion.Calculate_Ball_Motion
           (From          => (My_Position with Initial_Ball_Height),
            To            => Local_Destination,
            Initial_Speed => Initial_Speed,
            Angle         => Angle);

         Status.Manager.Manager.Throw(Team_Id        => Tid,
                                      Player_Task_Id => Pid,
                                      Motion         => Ball_Motion,
                                      Category => Category,
                                      Mode => Mode);
         Ball_Possession := False;
         Moving_Somewhere := False;
         Next_Execution :=
           Status.Get_Corresponding_Quantum(Init_Time + Throw_Time);
         Should_Release := False;
      end Throw;

      procedure Shot is
         X : Integer;
         Y : Integer;
      begin
         if Side = Left then
            X := Pitch.Right + 1;
         else
            X := Pitch.Left - 1;
         end if;
         Y := Field.Bottom_Touchline/2;
         Throw((X, Y, 0.0), Ada.Numerics.Pi/18.0);
      end Shot;


      procedure Play_The_Ball is
         Other_Team: Team_Id_T;
         Other_Side: Side_T;
         Opponents: Natural := 0;
         Teammate_Position : Coordinates_T;
      begin
         -- Initializing Other_* variables
         if Side = Left then Other_Side := Right;
         else Other_Side := Left; end if;
         if My_Id.Team_Id = Local then Other_Team := Visitor;
         else Other_Team := Local; end if;

         -- Read Players' Information
         Read_Players_Info(Players_Info, My_Position,
                           Motion.Get_Current_Position(Ball_Status.Motion));

         -- Calculate nearby opponents
         for i in Player_Index_T loop
            if Distance(My_Position, Players_Info.Positions(Other_Team, i)) < 3.0
            then
               Opponents := Opponents + 1;
            end if;
         end loop;

         -- Check whether the player should shot the ball into the goal
         if Field.Contains(Field.Penalty_Area(Other_Side), My_Position)
         then
            Shot;
            Next_Execution :=
              Status.Get_Corresponding_Quantum(Init_Time + Throw_Time);

         else
            Teammate_Position := My_Position;
            if not Field.Contains(Formation.Player_Zones(Pid), My_Position)
              or Current_Step >= 20 then -- not a good criteria, it's just enough
               Look_For_Teammate(Teammate_Position, Game_State.State);
            end if;
            if Teammate_Position /= My_Position then -- should pass
               Throw((Teammate_Position with 0.0));
            elsif Moving_Somewhere then
               Make_Next_Step;
            else
               Calculate_Movement;
            end if;
         end if;
      end Play_The_Ball;

      procedure Play_The_Game is
         Ball_Possessor : Identifier_T;
         Ball_Position : Coordinates_3D_T;
         Old_Ball_Status : Ball_Status_T;
      begin
         -- *** MUST HAVE FIELD CONTROL BEFORE READING ***
         -- READ the ball status keeping track of the old one
         Old_Ball_Status := Ball_Status;
         Ball_Status := Status.Ball_Status.Get;
         Ball_Possessor := Ball.Get_Possessor(Ball_Status);
         Ball_Possession :=Ball_Possessor = My_Id and not Ball_Status.Free;

         -- Check if the ball status has changed since the last execution
         Ball_Status_Changed := Ball_Status.Free /= Old_Ball_Status.Free
           or (Ball_Status.Possessor /= Old_Ball_Status.Possessor);

         -- Penalty for the previously tackled player (or one whom might
         -- have lost ball possession)
         if Old_Ball_Status.Possessor = My_Id
           and not Old_Ball_Status.Free
           and not Ball_Status.Free
           and Ball_Possessor.Team_Id /= My_Id.Team_Id
         then
            Next_Execution:=
              Status.Get_Corresponding_Quantum(Next_Execution + Penalty_Time);
            Moving_Somewhere := False;
            -- if the player has the ball possession...
         elsif Ball_Possession then
            Play_The_Ball;
         else	-- ...otherwise...
            -- Look at the current ball position
            Ball_Position := Motion.Get_Current_Position(Ball_Status.Motion);

            -- Check if the player can CATCH the ball
            if My_Position = Coordinates_T'(Ball_Position.X, Ball_Position.Y)
            then
               if Ball_Position.Z <= Player_Characteristics.Height + Player_Characteristics.Jump_Skill
               then
                  Catch;
               else
                  Next_Execution :=
                    Status.Get_Corresponding_Quantum(Init_Time + Catch_Time);
                  Utils.Logger.Add_Log
                    ("Player " & My_Id.Team_Id'Img & My_Id.Player_Number'Img
                     & " doesn't catch the ball because it's too high: "
                     & Ball_Position.Z'Img & "m");
               end if;
               -- Otherwise check if he can perform a TACKLE
            elsif Ball_Status.Free = False -- an opponent owns the ball
              and Ball_Status.Possessor.Team_Id /= My_Id.Team_Id
              and Field.Contains(Area => Field_Area_T'
                                   (My_Position.X-1, My_Position.X+1,
                                    My_Position.Y-1, My_Position.Y+1),
                                 Position => Ball_Position)
              and Status.Players_Status(Ball_Status.Possessor.Team_Id,
                                        Ball_Status.Possessor.Player_Number).Get_Role /= 1
            then -- TACKLE the opponent
               Tackle;
            elsif (not Moving_Somewhere) or else Ball_Status_Changed then
               -- if the player was not moving, or otherwise if he was
               -- moving but the ball status changed in the meantime
               Calculate_Movement;
            else
               Make_Next_Step;
            end if;
         end if;
      end;

      procedure Handle_Player_Condition is
      begin
         -- ***** PART I : PLAYER CONDITION HANDLING *****
         -- If there is a player notification or condition to handle
         -- (i.e. he has been injured, substituted or sent off), this should
         -- be done BEFORE any game interruption.
         if Interruption_Handling_Phase /= Player_Condition then
            Safely_Leave_Field_Control;
            -- *** 1st PHASE ***
            -- If the player hasn't yet started to handle his own condition,
            -- start leaving the field...
            Ball_Status := Ball_Status_T'
              (Motion            => Motion.Get_Fixed_Motion
                 (Game_Control.Game_State.Get.Position),
               Possessor         => Identifier_T'(Local,1), -- not relevant
               Free              => True);
            -- If substitutions are assigned at the end of the first period,
            -- the player should be aware that his field side has changed
            Side := Status.Teams_Status(Tid).Get_Side;
            Calculate_Movement;
            -- Remember (for the next task loops) that the player
            -- has already started the interruption handling...
            Interruption_Handling_Phase := Player_Condition;
         elsif Moving_Somewhere then
            -- *** 2nd PHASE ***
            -- Note: Handling_Interruption is True
            Make_Next_Step;
         else
            Safely_Leave_Field_Control; -- won't change position during this step
            -- *** 3rd PHASE ***
            -- Handling_Interruption is still True
            -- This branch is executed AFTER the player has left the field
            -- The action of leaving the field is done outside this procedure
            case Player_State is
               when Injured =>
                  -- Note Wait_For_Restart is not needed here
                  Status.Teams_Status(Tid).Set_Role_Status(Pid, False);
                  Game_Control.Game_State.Decrease_Players_Number;
                  select
                     -- ...but be prepared to stop waiting in case of
                     -- substitution (the player will then change his own
                     -- identity to restart as a new player)
                     Status.Players_Status(Tid, My_Id.Player_Number)
                       .Wait_For_Substitution;
                     Init_Time := Status.Get_Corresponding_Quantum(Clock);
                     -- Teams_Status(Tid).Set_Role_Status(Pid, True);
                     Game_Control.Game_State.Increase_Players_Number;
                  then abort
                     -- wait until the game is over...
                     Game_Control.Game_State.Wait_The_End;
                     Match_Finished := True;
                     Init_Time := Status.Get_Corresponding_Quantum(Clock);
                  end select;
               when Substitution =>
                  -- create player substitution event and add it to the buffer
                  Events_Buffer
                    .Add(
                      Player_Substitution_Event_T'(
                        Incoming_Player_Id =>
                          (My_Id.Team_Id,
                           Status.Players_Status(Tid, My_Id.Player_Number)
                           .Incoming_Substitute),
                        Outgoing_Player_Id => My_Id));
                  -- Update the local copy of the shirt number
                  My_Id.Player_Number :=
                    Status.Players_Status(Tid, My_Id.Player_Number)
                    .Incoming_Substitute;
                  -- update the global status of the player
                  Status.Player_Mapper(Tid,Pid).Set_Numbers
                    (My_Id.Player_Number,
                     Status.Players_Characteristics(Tid,My_Id.Player_Number).Shirt_Number);

                  Status.Players_Status(Tid, My_Id.Player_Number).Set_State(Active);
                  Status.Teams_Status(Tid).Set_Role_Status(Pid, True);
                  Player_State := Active;	-- maybe not mandatory
                  -- notify the frame manager
                  Television.Frame_Manager_Task.Substitute_Player
                    (Player_Identifier => My_Id,
                     Player_Task_Id    => Pid);
               when Sent_Off =>
                  -- Note Wait_For_Restart is not needed here
                  Status.Teams_Status(Tid).Set_Role_Status(Pid, False);
                  Game_Control.Game_State.Decrease_Players_Number;
                  Match_Finished := True;
                  -- Game_Control.Game_State.Wait_The_End;
               when others => -- the impossible happened... (active/inactive)
                  null;
            end case;
            -- The player condition handling is done
            Interruption_Handling_Phase := Not_Handling;
         end if; -- not handling interruption

      end Handle_Player_Condition;

      procedure Handle_Game_Interruption is
         Teammate_Position : Coordinates_T;
      begin
         Player_State := Status.Players_Status(Tid, My_Id.Player_Number).Get_State;
         if Interruption_Handling_Phase /= Game_Interruption
           or else Former_Game_State /= Game_State
         then
            -- *** 1st PHASE -- think ***
            Safely_Leave_Field_Control;
            Interruption_Handling_Phase := Game_Interruption;
            Ball_Status := Status.Ball_Status.Get;
            Calculate_Movement;
         elsif Moving_Somewhere then
            -- *** 2nd PHASE - move ***
            Make_Next_Step;
         else
            -- *** 3rd PHASE - restart ***
            Safely_Leave_Field_Control;
            -- The player is ready to start
            if Game_State.State = Half_Time_Break then
               Game_Control.Game_State.Wait_For_Positioning;
               Init_Time := Status.Get_Corresponding_Quantum(Clock);
            elsif Game_State.State = End_Match then
               Match_Finished := True;
            else
               Game_Control.Game_State.Wait_For_Restart;
               Init_Time := Status.Get_Corresponding_Quantum(Clock);

               if Status.Ball_Status.Get.Possessor.Team_Id = My_Id.Team_Id
                 and Status.Teams_Status(My_Id.Team_Id).Get_Starting_Player = Pid
               then
                  if Game_State.State = Penalty_Kick then
                     Shot;
                  else
                     Look_For_Teammate(Teammate_Position, Game_State.State);
                     Throw((Teammate_Position with 0.0));
                  end if;
               end if;
            end if;
            Interruption_Handling_Phase := Not_Handling;

            -- no need to set a delay because the player will be awaken
            -- by the referee at the beginning of a new quantum
            -- During his next loop the player will choose what to do.
         end if;
      end Handle_Game_Interruption;

      procedure Handle_Interruption is
      begin
         -- Verify the personal status of the player (active, injured, ...)
         Player_State :=
           Status.Players_Status(Tid, My_Id.Player_Number).Get_State;
         -- If he is not active, handle his personal position
         if Player_State /= Active then
            Handle_Player_Condition;
         else
            -- handle the interruption according to the referee's signal
            Handle_Game_Interruption;
         end if;
      end;

      ---------- TASK BODY STARTS HERE ----------
   begin
      -- Positioning can start only after the player characteristics have been set
      Game_Control.Game_State.Wait_For_Positioning;
      -- Initialize the player local data
      Initialize;

      -- During the first loop, the player will position himself to start playing
      -- (i.e. he reach his kick-off position, as said by the referee)
      while not Match_Finished loop
         -- Lock the player's neighborhood
         Get_Field_Control;

         -- Read Game state
         Former_Game_State := Game_State;
         Game_State := Game_Control.Game_State.Get;

         -- Start checking if the game was aborted
         if Game_State.Status = Match_Aborted then
            Match_Finished := True;
            -- if the game was not interrupted, play.
         elsif Game_State.State = Play then
            Play_The_Game;
         else
            Handle_Interruption;
         end if;
         Safely_Leave_Field_Control;
         delay until Next_Execution;
      end loop;

      -- everybody wait so that the television 'understands' when all the players
      -- have left the field (because television task uses Wait_For_Players).
      delay 1.0;
      if (Status.Players_Status(Tid,My_Id.Player_Number).Get_State /= Sent_Off
          or else Interruption_Handling_Phase /= Not_Handling)
        and (Status.Players_Status(Tid,My_Id.Player_Number).Get_State /= Injured
             or else Interruption_Handling_Phase /= Not_Handling)
      then
         Game_Control.Game_State.Decrease_Players_Number;
      end if;
   end Player_Task_T;
end Player.Logic;
