with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Output_Types; use Output_Types;

package Player is
   -- Given a team, return the name of the other one
   function Opposite_Team (Team : Team_Id_T) return Team_Id_T;

   -- a record type to represent a player identifier
   type Identifier_T is record
      Team_Id : Team_Id_T;
      Player_Number : Player_Number_T;
   end record;

   --+---------------------------
   --  THE PLAYER CHARACTERISTICS
   --+---------------------------
   -- For each player, provide his static characteristics
   type Player_Characteristics_T is record
      Name     		: Unbounded_String;
      Shirt_Number	: Positive;
      Speed       	: Float := 4.0;
      Stamina     	: Float range 0.0..1.0 := 0.8;
      Passing_Power  	: Float := 20.0;	-- m/s
      Passing_Accuracy 	: Float range 0.0..1.0 := 0.8;
      Shooting_Power  	: Float := 30.0;	-- m/s
      Shooting_Accuracy : Float range 0.0..1.0 := 0.8;
      Height      	: Float := 1.8;		-- m
      Jump_Skill       	: Float := 0.5;		-- m
      Heading_Power	: Float := 15.0;	-- m/s
      Heading_Accuracy	: Float range 0.0..1.0 := 0.8;
      Dribble_Accuracy  : Float range 0.0..1.0 := 0.8;
      Tackle_Skill      : Float range 0.0..1.0 := 0.8;
      Catching_Accuracy	: Float range 0.0..1.0 := 0.9;
      Aggression  	: Float range 0.0..1.0 := 0.5;
      Teak_Work_Skill  	: Float range 0.0..1.0;
   end record;

   -- *** PLAYER STATUS ***
   -- For each player, provide his dynamic characteristics, i.e. current fitness,
   -- status (active, injured, substituted, ...)
   protected type Player_Status_T is
      -- a procedure to decrease the player fitness (used in tackle handling)
      procedure Decrease_Fitness_Value(Damage 	 : in Float;
                                       With_Foul : in Boolean;
                                       Caused_Injury: out Boolean;
                                       New_Fitness : out Float);

      -- overloading to decrease player fitness in other circustamces (fatigue)
      procedure Decrease_Fitness_Value(Damage 	 : in Float;
                                       New_Fitness : out Float);

      -- a procedure to set a player under substitution
      procedure Substitute (Substituted_Id : in Player_Number_T);

      -- Let an injured player wait to be substituted
      entry Wait_For_Substitution;

      -- Read the shirt number of the substitute
      function Incoming_Substitute return  Player_Number_T;

      -- Change the current status of a player (may be injured, sent off, ...)
      procedure Set_State(New_State: in Player_State_T);

      -- Change the role of a player into the current formation (11 roles)
      procedure Set_Role(New_Role: in Player_Index_T);

      -- Return the fitness value of a player
      function Get_Fitness_Value return Float;

      -- Return the current status of a player
      function Get_State return Player_State_T;

      -- Return the role ID of a player in the current team formation
      function Get_Role return Player_Index_T;
   private
      State : 	Player_State_T := Inactive;
      Fitness :	Float := 100.0; -- 0 means that the player is injured
      -- the player role in the current formation; it represent an index for
      -- the Player_Standard_Positions_T array
      Role :	Player_Index_T;
      -- the shirt number of substituting player (used only during substitutions)
      Substituted_By : Player_Number_T;
   end Player_Status_T;

   -- *** "PLAYER MAPPER" ***
   -- a protected type to map a player task id to his current shirt number
   -- and position inside the field. Having the position here is useful because
   -- it permits to check all the active players' position without checking
   -- also the player status of those that are not playing
   protected type Player_Mapper_T is
      -- getters and setters
      procedure Set_Numbers (New_Player_Number : in Player_Number_T;
                             New_Shirt_Number  : in Positive);
      procedure Set_Position (New_Position : in Coordinates_T);
      function Get_Player_Number return Player_Number_T;
      function Get_Shirt_Number return Positive;
      function Get_Position return Coordinates_T;
   private
      -- the player t-shirt number
      Shirt_Number : Positive;
      Player_Number : Player_Number_T;
      -- the current player position
      Position : Coordinates_T;
   end Player_Mapper_T;
end Player;
