with Output_Types; use Output_Types;
with Status;
with Ball;
with Position;
with Motion;
with Field; use Field;
with Referee; use Referee;
with Player;
with Utils;
with Events; use Events;
with Events.Ball_Out; use Events.Ball_Out;

package body Assistant_Referee is
   task body Assistant_Referee_Task is
      -- a boolean to store the assistant referee status (active or asleep);
      -- when active the task is waiting the ball to leave the field
      Active : Boolean := False;
      Match_Finished : Boolean := False;
      -- the time when the ball will (potentially) leave the the field
      Ball_Out_Time : Time;
   begin
      while not Match_Finished loop
         select
            accept Set_Ball_Out_Time (New_Time : in Time) do
               -- set the leaving time of the ball and
               -- activate the assistant referee
               Ball_Out_Time := New_Time;
               Active := True;
            end Set_Ball_Out_Time;
         or
            accept Reset_Ball_Out_Time  do
               -- deactivate the assistant referee
               Active := False;
            end Reset_Ball_Out_Time;
         or
            accept Quit do
               Match_Finished := True;
            end Quit;
         or
            when Active =>
               delay until Ball_Out_Time;
               -- being here means the ball is out of the field:
               declare
                  -- local var to store current ball status
                  Ball_Status : constant Ball.Ball_Status_T := Status.Ball_Status.Get;
                  -- local var to store current ball position
                  Ball_Position : constant Coordinates_3D_T :=
                    Motion.Element_At(Ball.Get_Motion(Ball_Status),
                                      Motion.Last(Ball_Status.Motion));
                  -- local var to store the identifier of the player which own the ball
                  Last_Touch_Player_Id : constant Player.Identifier_T :=
                    Ball.Get_Possessor(Ball_Status);
                  -- local var to store the identifier of the team which own the ball
                  Last_Touch_Team_Id : constant Team_Id_T :=
                    Ball.Get_Possessor(Ball_Status).Team_Id;
                  -- local var to store the side of the team which own the ball
                  Last_Touch_Team_Side : constant Field.Side_T :=
                    Status.Teams_Status(Last_Touch_Player_Id.Team_Id).Get_Side;
               begin
                  -- check which line the ball has crossed
                  if Ball_Position.X = Position.X_Range_T'First then
                     -- left goal line, check if the ball is in goal
                     if Ball_Position.Y <= Field.Goal_Y_Max and
                       Ball_Position.Y >= Field.Goal_Y_Min and
                       Ball_Position.Z < Field.Crossbar_Height
                     then
                        -- goal (yay!), check if it is a goal or an autogoal
                        if Last_Touch_Team_Side = Right then
                           -- goal, create the ball out (goal) event and add it
                           -- to the buffer
                           Events_Buffer
                             .Add(Ball_Out_Event_T'(
                               Category => Goal,
                               Which_Goal => Player.Opposite_Team(Last_Touch_Player_Id.Team_Id)));
                           -- notify the referee passing the id of the team
                           -- which has made the goal
                           Referee_Task.Goal(Last_Touch_Player_Id,
                                             Last_Touch_Player_Id.Team_Id,
                                             False);
                        else
                           -- autogol, create the ball out (goal) event and add
                           -- it to the buffer
                           Events_Buffer
                             .Add(Ball_Out_Event_T'(
                               Category => Goal,
                               Which_Goal => Last_Touch_Team_Id));
                           -- notify the referee passing the id of the team
                           -- wich has gained the goal
                           Referee_Task.Goal(
                             Last_Touch_Player_Id,
                             Player.Opposite_Team(Last_Touch_Player_Id.Team_Id),
                             True);
                        end if;
                     else
                        -- ball out of field (from left goal line)
                        -- check if it is a corner kick or a goal kick
                        if Last_Touch_Team_Side = Right then
                           -- goal kick, create the ball out (goal line) event
                           -- and add it to the buffer
                           Events_Buffer.
                             Add(Ball_Out_Event_T'(Category => Goal_Line,
                                                   Which_Goal => Last_Touch_Team_Id));
                           -- notify the referee passing the id of the team which
                           -- has sent off the ball
                           Referee.Referee_Task.Goal_Kick(Last_Touch_Team_Id);
                        else
                           -- corner kick, create the ball out event and add it
                           -- to the buffer
                           Events_Buffer.
                             Add(Ball_Out_Event_T'(Category => Goal_Line,
                                                   Which_Goal => Last_Touch_Team_Id));
                           -- notify the referee passing the ball status
                           -- which contains the ball position and the id of the team
                           -- which has touched the ball last
                           Referee.Referee_Task.Corner_Kick(Ball_Status);
                        end if;
                     end if;
                  elsif Ball_Position.X = Position.X_Range_T'Last then
                     -- right goal line, check if the ball is in goal
                     if Ball_Position.Y <= Field.Goal_Y_Max and
                       Ball_Position.Y >= Field.Goal_Y_Min and
                       Ball_Position.Z < Field.Crossbar_Height
                     then
                        -- goal (yay!), check if it is an autogoal
                        if Last_Touch_Team_Side = Left then
                           -- goal, create the ball out (goal) event and add it
                           -- to the buffer
                           Events_Buffer
                             .Add(Ball_Out_Event_T'(
                               Category => Goal,
                               Which_Goal => Player.Opposite_Team(Last_Touch_Team_Id)));
                           -- notify the referee passing the id of the team
                           -- which has made the goal
                           Referee_Task.Goal(Last_Touch_Player_Id,
                                             Last_Touch_Player_Id.Team_Id,
                                             False);
                        else
                           -- autogol, create the ball out (goal) event and add
                           -- it to the buffer
                           Events_Buffer
                             .Add(Ball_Out_Event_T'(
                               Category => Goal,
                               Which_Goal => Last_Touch_Team_Id));
                           -- notify the referee passing the id of the team
                           -- wich has gained the goal
                           Referee_Task.Goal(
                             Last_Touch_Player_Id,
                             Player.Opposite_Team(Last_Touch_Player_Id.Team_Id),
                             True);
                        end if;
                     else
                        -- ball out of field (from right goal line)
                        -- check if it is a corner kick or a goal kick
                        if Last_Touch_Team_Side = Left then
                           -- goal kick, create the ball out (goal line) event
                           -- and add it to the buffer
                           Events_Buffer
                             .Add(Ball_Out_Event_T'(Category => Goal_Line,
                                                    Which_Goal => Last_Touch_Team_Id));
                           -- notify the referee passing the id of the
                           -- team which has sent off the ball
                           Referee.Referee_Task.Goal_Kick(Last_Touch_Team_Id);
                        else
                           -- corner kick, create the ball out (goal line) event
                           -- and add it to the buffer
                           Events_Buffer
                             .Add(Ball_Out_Event_T'(Category => Goal_Line,
                                                    Which_Goal => Last_Touch_Team_Id));
                           -- notify the referee passing the ball status
                           -- which contains the ball position and the id of the
                           -- team which has touched the ball last
                           Referee.Referee_Task.Corner_Kick(Ball_Status);
                        end if;
                     end if;
                  else
                     -- touch line, create the ball out (touch line) event and
                     -- add it to the buffer
                     Events_Buffer
                       .Add(Ball_Out_Event_T'(Category => Touch_Line,
                                              Which_Goal => Last_Touch_Team_Id));
                     -- notify the referee passing the ball status
                     -- which contains the ball position and the id of the team
                     -- which has touched the ball last
                     Referee.Referee_Task.Throw_In(Ball_Status);
                  end if;
               end;
               Active := False;
         end select;
      end loop;
   end Assistant_Referee_Task;
end Assistant_Referee;
