with Status; use Status;
with Field_Control;
with Game_Control; use Game_Control;
with Motion;
with Field; use Field;
with Ada.Containers.Doubly_Linked_Lists; use Ada.Containers;
with Utils;
with Assistant_Referee;
with Television; use Television;
with Tactic;
with Events; use Events;
with Events.Goal; use Events.Goal;
with Events.Interruption; use Events.Interruption;
with Events.Kick; use Events.Kick;
with Events.Penalty_Card; use Events.Penalty_Card;
with Events.Restart; use Events.Restart;
with Events.Start_Match; use Events.Start_Match;
with Events.Substitution; use Events.Substitution;
with Events.Substitution_Request; use Events.Substitution_Request;
with Events.Time_Out; use Events.Time_Out;
with Events.Termination; use Events.Termination;
with Events.Abort_Match; use Events.Abort_Match;
with Events.State_Change; use Events.State_Change;

package body Referee is
   -- a constant to store the duration of a period
   Period_Duration : Time_Span;
   -- a constant to store the duration of the half time break
   Half_Time_Break_Duration : Time_Span;
   -- a constant to store the max num of substitutions allowed
   Substitutions_Limit : Natural;
   -- a constant to store the tackle damage limit for cards
   Yellow_Card_Limit : constant := 60.0;
   Red_Card_Limit : constant := 80.0;
   -- an enum type to represent a sanction card
   type Card_T is (None, Yellow, Red);
   -- we istantiate the Doubly_Linked_Lists generic to implement a list of
   -- substitution requests
   package Substitutions_Request_List is
     new Ada.Containers.Doubly_Linked_Lists(Substitution_Request_T);

   procedure Swap_Teams_Sides is
      New_Player_Position : Coordinates_T;
      Current_Player_Status: Player_State_T;
   begin
      -- swap the teams in the game field
      Status.Teams_Status(Output_Types.Local).Swap_Side;
      Status.Teams_Status(Output_Types.Visitor).Swap_Side;
      for t in Team_Id_T loop
         for p in Player_Index_T loop
            Current_Player_Status :=
              Status.Players_Status(t,Status.Player_Mapper(t,p).Get_Player_Number).Get_State;
            if Current_Player_Status = Injured or Current_Player_Status = Sent_Off
            then
               New_Player_Position :=
                 Tactic.Outside_Position(Side => Status.Teams_Status(t).Get_Side,
                                         Role_Index => p);
               Status.Player_Mapper(t,p).Set_Position(New_Player_Position);
               Television.Frame_Manager_Task.Move_Player(t,
                                                         p,
                                                         New_Player_Position);
            end if;
         end loop;
      end loop;
   end Swap_Teams_Sides;

   task body Referee_Task is

      -- a var to store the id of the team which has to start the match
      Starter_Team : constant Team_Id_T := Output_Types.Local;
      -- a boolean to store if the current period is ended
      Current_Period_Is_Ended : Boolean := False;
      -- an enum to store the referee local match status
      Current_Status : Output_Types.Match_Status := Game_State.Get.Status;
      -- a time var to store the referee time out value corresponding to the end
      -- of the current period or the end of the half time break (depending on
      -- the referee state); it is initialized with the first period end time
      Time_Out : Time := Status.Get_Corresponding_Quantum(Status.Start_Time +
                                                            Period_Duration);
      -- a bidimensional array to store the cards registry
      Cards_Registry : array (Team_Id_T,Player_Number_T) of Card_T :=
        (others => (others => None));
      -- two lists (one for each team) to store pending substitution requests
      Substitutions_Registry : array (Team_Id_T) of Substitutions_Request_List.List;
      -- two natural (one for each team) to store the number of substitutions made
      Substitutions_Num : array (Team_Id_T) of Natural := (0 , 0);
      -- local var to store the ball status
      Local_Ball_Status : Ball.Ball_Status_T;
      -- local var to store the team id
      Local_Team_Id : Team_Id_T;
      -- local var to store the player id
      Local_Player_Id : Player.Identifier_T;
      -- local var to store the player position
      Local_Player_Position : Coordinates_T;
      -- local var to store the tackle damage
      Local_Tackle_Damage : Float;
      -- local boolean to store if the foul is with injury
      Local_Caused_Injury : Boolean;
      -- local boolean to say if a goal is an autogoal
      Local_Is_Autogoal : Boolean;

      -- a (local) function to check if a substitution is legal
      function Check_Admissibility (Substitution_Request : in Substitution_Request_T)
                                    return Boolean is
         -- local vars to store the result, the team id and the outgoing/incoming
         -- players state
         Success : Boolean := False;
         Team_Id : constant Team_Id_T :=
           Substitution_Request.Team_Id;
         Outgoing_Player_State: constant Player_State_T :=
           Status.Players_Status(Team_Id,
                                 Status.Shirt_To_Player_Number(Team_Id).Element
                                 (Substitution_Request.Outgoing_Shirt_Number))
           .Get_State;
         Incoming_Player_State : constant Player_State_T :=
           Status.Players_Status(Team_Id,
                                 Status.Shirt_To_Player_Number(Team_Id).Element
                                 (Substitution_Request.Incoming_Shirt_Number))
           .Get_State;
      begin
         -- substitution request is legal if:
         -- * the team has not reached the max number of substitutions allowed
         -- * the outgoing player is not already substituted, previously sent off or
         --   already out of game field
         -- * the incoming player is not already substituted, previously sent off or
         --   already in the game field (and not injured)
         -- and not legal otherwise
         if Substitutions_Num(Team_Id) < Substitutions_Limit
           and (Outgoing_Player_State = Output_Types.Active or
                  Outgoing_Player_State = Output_Types.Injured)
           and Incoming_Player_State = Output_Types.Inactive
         then
            Success := True;
         end if;
         return Success;
      end Check_Admissibility;

      -- a (local) procedure to handle pending substitution requests for a team
      procedure Handle_Substitutions (Team_Id : in Team_Id_T) is
      begin
         -- while the substitution registry is not empty, handle requests
         while not Substitutions_Registry(Team_Id).Is_Empty loop
            -- get the first substitution request from registry
            declare
               -- local (itaration) var to store the current substitution request
               -- and the incoming/outgoing players shirt number
               Request : constant Substitution_Request_T :=
                 Substitutions_Registry(Team_Id).First_Element;
               Incoming_Player_Id : constant Positive :=
                 Status.Shirt_To_Player_Number(Team_Id).Element(Request.Incoming_Shirt_Number);
               Outgoing_Player_Id : constant Positive :=
                 Status.Shirt_To_Player_Number(Team_Id).Element(Request.Outgoing_Shirt_Number);
            begin
               -- check if the substitution is legal
               -- TODO: this code is repeated, move into a local function
               if Check_Admissibility(Request) then
                  -- create the substitution event and push it into the buffer
                  Events_Buffer
                    .Add(Substitution_Event_T'(
                      Incoming_Player_Id => (Team_Id,Incoming_Player_Id),
                      Outgoing_Player_Id => (Team_Id,Outgoing_Player_Id),
                      Agreed             => True));
                  -- substitute player
                  Status.Players_Status(Team_Id,Outgoing_Player_Id)
                    .Substitute(Incoming_Player_Id);
                  -- increment the number of substitution made by team
                  Substitutions_Num(Team_Id) := Substitutions_Num(Team_Id) + 1;
               else
                  -- create the substitution event and push it into the buffer
                  Events_Buffer
                    .Add(Substitution_Event_T'(
                      Incoming_Player_Id => (Team_Id,Incoming_Player_Id),
                      Outgoing_Player_Id => (Team_Id,Outgoing_Player_Id),
                      Agreed             => False));
               end if;
            end;
            -- substitution handled, delete it from registry
            Substitutions_Registry(Team_Id).Delete_First;
         end loop;
      end Handle_Substitutions;

      -- a (local) procedure to handle the end of the current period
      procedure Handle_End_Period is
         Match_Status : Output_Types.Match_Status := Game_State.Get.Status;
      begin
         -- check the match status
         case Match_Status is
            when Match_Started =>
               -- we are during the first half
               -- create the first half time out (with whistle) event and push it
               -- into the buffer
               Events_Buffer.Add(First_Half_Time_Out_Event_T'(Is_Whistle => True));
               -- handle first period ending, set the new referee state
               Current_Period_Is_Ended := False;
               Current_Status := First_Half_Finished;
               -- stop the television timer
               Television.Frame_Manager_Task.Stop_Timer;
               -- swap the team
               Swap_Teams_Sides;
               -- set the new game state
               Game_State.Half_Time_Break;
               -- wait for player to go out
               Game_State.Enable_Start_2nd_Time_Request;
               -- set the new time out value to the end of the half time break
               Time_Out :=
                 Status.Get_Corresponding_Quantum(Clock + Half_Time_Break_Duration);
               -- create the state change event and push it into the buffer
               Events_Buffer.Add(State_Change_Event_T'(State => Current_Status));
               -- restart the television timer
               Television.Frame_Manager_Task.Restart_Timer;
            when Second_Half_Started =>
               -- we are during the second half
               -- create the first half time out (with whistle) event and push it
               -- into the buffer
               Events_Buffer.Add(Second_Half_Time_Out_Event_T'(Is_Whistle => True));
               -- handle second period ending
               Current_Status := Second_Half_Finished;
               -- stop the television timer
               Television.Frame_Manager_Task.Stop_Timer;
               -- create the state change event and push it into the buffer
               Events_Buffer.Add(State_Change_Event_T'(State => Current_Status));
               -- set the new game state
               Game_State.End_Match;
            when others =>
               -- in any other status do nothing
               null;
         end case;
      end Handle_End_Period;

      -- a (local) procedure to start the second period after the end of the
      -- half time break (for time out or after an explicit request)
      procedure Start_Second_Period is
         -- local var to store the new ball position
         New_Ball_Position   : constant Coordinates_3D_T :=
           Coordinates_3D_T'(Field.Midfield(Right) with 0.0);
         -- local var to store the id of the team which has to restart the game
         Offence_Team_Id     : constant Team_Id_T :=
           Player.Opposite_Team(Starter_Team);
         -- local var to store the new game state
         New_Game_State      : constant State_Record_T :=
           State_Record_T'(Game_Control.Kick_Off,
                           Second_Half_Started,
                           New_Ball_Position,
                           Offence_Team_Id);
         -- local vars to store the new ball status
         New_Ball_Motion     : constant Motion.Motion_3d_T :=
           Motion.Get_Fixed_Motion(New_Ball_Position);
         Current_Ball_Status : constant Ball.Ball_Status_T :=
           Status.Ball_Status.Get;
         New_Ball_Status     : constant Ball.Ball_Status_T :=
           (New_Ball_Motion,
            Ball.Get_Possessor(Current_Ball_Status),
            False);
      begin
         -- stop the television timer
         Television.Frame_Manager_Task.Stop_Timer;
         -- set the starting player of the offence team
         Status.Teams_Status(Offence_Team_Id)
           .Set_Starting_Player(New_Game_State);
         -- set the new game state
         Game_State.Set(New_Game_State);
         -- update the referee current status
         Current_Status := Second_Half_Started;
         -- The ball can be positioned in advance because players are all out
         -- of the field.
         Status.Ball_Status.Set(New_Ball_Status);
         -- here all *active* players are ready to restart the game
         -- update the ball status
         Game_State.Wait_For_Players;
         -- Players held during the 1st time period are now released
         -- wait for players
         -- create the state change event and push it into the buffer
         Events_Buffer.Add(State_Change_Event_T'(State => Second_Half_Started));
         -- create the restart event and push it into the buffer
         Events_Buffer.Add(Restart_Event_T'(null record));
         delay until Status.Get_Corresponding_Quantum(Clock + Status.Quantum);
         -- restart the television timer
         Television.Frame_Manager_Task.Restart_Timer;
         Game_State.Start;
         -- set the time out value to the second period end time
         Time_Out := Clock + Period_Duration;
      end Start_Second_Period;

      procedure Get_New_Ball_Status(New_Ball_Motion : in Motion.Motion_3d_T;
                                    Offence_Team_Id : in Team_Id_T;
                                    New_Ball_Status : out Ball.Ball_Status_T)
      is
      begin
         New_Ball_Status :=
           Ball.Ball_Status_T'
             (Motion => New_Ball_Motion,
              Possessor => Identifier_T'
                (Team_Id => Offence_Team_Id,
                 Player_Number => Status.Player_Mapper
                   (Offence_Team_Id,
                    Status.Teams_Status(Offence_Team_Id)
                    .Get_Starting_Player).Get_Player_Number),
              Free => False);
      end Get_New_Ball_Status;
   begin
      -- Initialize the game
      declare
         Init_State_Record : State_Record_T :=
           (State        => Not_Started,
            Status	 => Warm_Up,
            Position     => Coordinates_3D_T'(Midfield(Left) with 0.0),
            Offence_Team => Local);
      begin
         -- Initialize Game_State
         Game_State.Set(Init_State_Record);
         -- prepare the new state record
         Init_State_Record.State := Kick_Off;
         -- ask the offence team to decide the starting player
         Status.Teams_Status(Init_State_Record.Offence_Team)
           .Set_Starting_Player(Init_State_Record);
         -- place the ball on the midfield spot (no need to wait the players here)
         Status.Ball_Status.Set_Position(Init_State_Record.Position);
         -- wait fot the game initialization
         accept Game_Configuration (New_Period_Duration: in Time_Span;
                                    New_Half_Time_Break_Duration: in Time_Span;
                                    New_Substitutions_Limit: in Positive) do
            Period_Duration := New_Period_Duration;
            Half_Time_Break_Duration :=  New_Half_Time_Break_Duration;
            Substitutions_Limit := New_Substitutions_Limit;
         end Game_Configuration;
         -- let the players positioning for the kick-off
         delay until Status.Get_Corresponding_Quantum(Clock + Status.Quantum);
         Current_Status := Match_Started;
         Init_State_Record.Status := Current_Status;
         -- create the state change event and push it into the buffer
         Events_Buffer.Add(State_Change_Event_T'(State => Current_Status));
         -- set the new game state
         Game_State.Set(Init_State_Record);
      end;
      -- wait the players
      Game_Control.Game_State.Wait_For_Players;
      -- create the start match event and push it into the buffer
      Events_Buffer.Add(Start_Match_Event_T'(null record));
      -- start the match
      delay until Status.Get_Corresponding_Quantum(Clock + Status.Quantum);
      -- restart the television timer
      Television.Frame_Manager_Task.Restart_Timer;
      Game_Control.Game_State.Start;
      Time_Out := Status.Get_Corresponding_Quantum(Clock + Period_Duration);
      -- referee main loop
      while Current_Status /= Second_Half_Finished
        and Current_Status /= Match_Aborted
      loop
         select
            -- handle foul interruption
            -- Player_Id is the identifier of the player who committed the foul
            accept Foul (Player_Id 		: in Player.Identifier_T;
                         Player_Position	: in Coordinates_T;
                         Tackle_Damage 		: in Float;
                         Caused_Injury		: in Boolean) do
               -- the randevouz is used only for parameter passing
               Local_Player_Id := Player_Id;
               Local_Player_Position := Player_Position;
               Local_Tackle_Damage := Tackle_Damage;
               Local_Caused_Injury := Caused_Injury;
            end Foul;
            -- create the interruption event and push it to the buffer
            Events_Buffer.Add(Interruption_Event_T'(Category => Foul));
            -- check if there is a sanction to assign
            if Local_Tackle_Damage >= Yellow_Card_Limit and
              Local_Tackle_Damage < Red_Card_Limit then
               -- assign yellow card; if player already has a yellow card, assign
               -- a red one
               if Cards_Registry(Local_Player_Id.Team_Id,
                                 Local_Player_Id.Player_Number) = Yellow then
                  -- create the penalty card event (second yellow) and push it
                  -- into the buffer
                  Events_Buffer
                    .Add(Penalty_Card_Event_T'(Player_Id => Local_Player_Id,
                                               Category  => Second_Yellow));

                  -- player already has a yellow card, assign a red one and send
                  -- him off
                  Cards_Registry(Local_Player_Id.Team_Id,
                                 Local_Player_Id.Player_Number) := Red;
                  Status.Players_Status(Local_Player_Id.Team_Id,
                                        Local_Player_Id.Player_Number)
                    .Set_State(Output_Types.Sent_Off);
               else
                  -- create the penalty card event (first yellow) and push it
                  -- into the buffer
                  Events_Buffer
                    .Add(Penalty_Card_Event_T'(Player_Id => Local_Player_Id,
                                               Category  => First_Yellow));

                  -- player has not a yellow card, assign it
                  Cards_Registry(Local_Player_Id.Team_Id,
                                 Local_Player_Id.Player_Number) := Yellow;
               end if;
            elsif Local_Tackle_Damage >= Red_Card_Limit then
               -- create the penalty card event (second yellow) and push it
               -- into the buffer
               Events_Buffer
                 .Add(Penalty_Card_Event_T'(Player_Id => Local_Player_Id,
                                            Category  => Red));

               -- assign red card and send player off
               Cards_Registry(Local_Player_Id.Team_Id,
                              Local_Player_Id.Player_Number) := Red;
               Status.Players_Status(Local_Player_Id.Team_Id,
                                     Local_Player_Id.Player_Number)
                 .Set_State(Output_Types.Sent_Off);
            end if;
            -- handle substitutions (if any) for each team
            Handle_Substitutions(Output_Types.Local);
            Handle_Substitutions(Output_Types.Visitor);
            declare
               -- local var to store the offended team identifier
               Offended_Team_Id : constant Team_Id_T :=
                 Player.Opposite_Team(Local_Player_Id.Team_Id);
               -- local var to store the side of the offended team
               Offended_Team_Side : constant Field.Side_T :=
                 Status.Teams_Status(Offended_Team_Id).Get_Side;
               -- local var to store the new ball position;
               New_Ball_Position : Coordinates_3D_T;
               -- local var to store the new game state value;
               New_Game_State_Value : Game_Control.State_T;
               -- local var to store the new game state;
               New_Game_State : State_Record_T;
               -- local vars to store the new ball status;
               New_Ball_Motion : Motion.Motion_3d_T;
               New_Ball_Status : Ball.Ball_Status_T;
            begin
               -- determine if the foul is a "penalty kick" or a "free kick" foul
               -- to get the new game state
               if Offended_Team_Side = Field.Right and
                 Field.Contains(Left_Penalty_Area, Local_Player_Position) then
                  -- create the penalty kick event and push it into the buffer
                  Events_Buffer.Add(Penalty_Kick_Event_T'(Team_Id => Offended_Team_Id));
                  -- penalty kick on the left side, new ball position is the left
                  -- penalty spot and the new state is penalty kick
                  New_Ball_Position :=
                    Coordinates_3D_T'(Field.Left_Penalty_Spot with 0.0);
                  New_Game_State_Value := Game_Control.Penalty_Kick;
               elsif Offended_Team_Side = Left and
                 Field.Contains(Right_Penalty_Area, Local_Player_Position) then
                  -- create the penalty kick event and push it into the buffer
                  Events_Buffer.Add(Penalty_Kick_Event_T'(Team_Id => Offended_Team_Id));
                  -- penalty kick on the right side, new ball position is the right
                  -- penalty spot and the new state is penalty kick
                  New_Ball_Position :=
                    Coordinates_3D_T'(Field.Right_Penalty_Spot with 0.0);
                  New_Game_State_Value := Game_Control.Penalty_Kick;
               else
                  -- create the free kick event and push it into the buffer
                  Events_Buffer.Add(Free_Kick_Event_T'(Team_Id => Offended_Team_Id));
                  -- free kick, new ball position is the position of the player
                  -- who committed the foul and the new state is free kick
                  New_Ball_Position :=
                    Coordinates_3D_T'(Local_Player_Position with 0.0);
                  New_Game_State_Value := Game_Control.Free_Kick;
               end if;
               -- set the new game state
               New_Game_State := State_Record_T'(New_Game_State_Value,
                                                 Current_Status,
                                                 New_Ball_Position,
                                                 Offended_Team_Id);
               Status.Teams_Status(Offended_Team_Id)
                 .Set_Starting_Player(New_Game_State);
               Game_State.Set(New_Game_State);
               -- deregister the player who committed the foul
               Field_Control.Barrier.Leave(Local_Player_Position);
               -- wait for players
               Game_State.Wait_For_Players;
               -- here all *active* players are ready to restart the game
               -- update the ball status; the new possessor is not set so it is
               -- the last player which touch the ball
               New_Ball_Motion := Motion.Get_Fixed_Motion(New_Ball_Position);
               Get_New_Ball_Status(New_Ball_Motion, Offended_Team_Id, New_Ball_Status);
               -- create the restart event and push it into the buffer
               Events_Buffer.Add(Restart_Event_T'(null record));
               Status.Ball_Status.Set(New_Ball_Status);
               -- set the new game state
               delay until Status.Get_Corresponding_Quantum(Clock + Status.Quantum);
               Game_State.Start;
            end;
         or
            -- handle throw-in interruption
            accept Throw_In (Ball_Status : in Ball.Ball_Status_T) do
               -- the randevouz is used only for parameter passing
               Local_Ball_Status := Ball_Status;
            end Throw_In;
            -- create the interruption event and push it into the buffer
            Events_Buffer.Add(Interruption_Event_T'(Category => Ball_Out));
            -- handle substitutions (if any) for each team
            Handle_Substitutions(Output_Types.Local);
            Handle_Substitutions(Output_Types.Visitor);
            -- check if current period is ended
            if Current_Period_Is_Ended then
               Handle_End_Period;
            else
               declare
                  -- local var to store the ball new position
                  New_Ball_Position : constant Coordinates_3D_T :=
                    Local_Ball_Status.Motion.Get_Current_Position;
                  -- local var to store the id of the team which has to restart game
                  Offence_Team_Id   : constant Team_Id_T :=
                    Player.Opposite_Team(Local_Ball_Status.Possessor.Team_Id);
                  -- local var to store the new game state
                  New_Game_State    : constant State_Record_T :=
                    State_Record_T'(Game_Control.Throw_In,
                                    Current_Status,
                                    New_Ball_Position,
                                    Offence_Team_Id);
                  -- local vars to store the new ball status
                  New_Ball_Motion   : constant Motion.Motion_3d_T :=
                    Motion.Get_Fixed_Motion(New_Ball_Position);
                  New_Ball_Status   : Ball.Ball_Status_T;
               begin
                  -- create the throw-in event and push it into the buffer
                  Events_Buffer.Add(Throw_In_Event_T'(Team_Id => Offence_Team_Id));
                  Status.Teams_Status(Offence_Team_Id).Set_Starting_Player(New_Game_State);
                  Get_New_Ball_Status(New_Ball_Motion, Offence_Team_Id, New_Ball_Status);
                  -- set the new game state
                  Game_State.Set(New_Game_State);
                  -- wait for players
                  Game_State.Wait_For_Players;
                  -- here all *active* players are ready to restart the game
                  -- create the restart event and push it into the buffer
                  Events_Buffer.Add(Restart_Event_T'(null record));
                  -- update the ball status
                  Status.Ball_Status.Set(New_Ball_Status);
                  -- set the new game state
                  delay until Status.Get_Corresponding_Quantum(Clock + Status.Quantum);
                  Game_State.Start;
               end;
            end if;
         or
            -- handle goal kick interruption
            accept Goal_Kick (Team_Id : in Team_Id_T) do
               -- the randevouz is used only for parameter passing
               Local_Team_Id := Team_Id;
            end Goal_Kick;
            -- create the interruption event and push it into the buffer
            Events_Buffer.Add(Interruption_Event_T'(Category => Ball_Out));
            -- handle substitutions (if any) for each team
            Handle_Substitutions(Output_Types.Local);
            Handle_Substitutions(Output_Types.Visitor);
            -- check if the current period is ended
            if Current_Period_Is_Ended then
               Handle_End_Period;
            else
               declare
                  -- local var to store the offence team id
                  Offence_Team_Id   : constant Team_Id_T :=
                    Player.Opposite_Team(Local_Team_Id);
                  -- local var to store the (current) ball position
                  Ball_Position     : constant Coordinates_3D_T :=
                    Status.Ball_Status.Get.Motion.Get_Current_Position;
                  -- local var to store the new ball position
                  New_Ball_Position : Coordinates_3D_T;
                  -- local var to store the new game state
                  New_Game_State    : State_Record_T;
                  -- local vars to store the new ball status
                  New_Ball_Motion   : Motion.Motion_3d_T;
                  New_Ball_Status   : Ball.Ball_Status_T;
                  Starting_Player   : Identifier_T;
               begin
                  -- check where the ball has left the field to get the new ball
                  -- position
                  if Ball_Position.X < Field.Left_Goal_Line then
                     -- ball sent off on left side of the field
                     if Ball_Position.Y > Field.Bottom_Touchline / 2 then
                        -- on the right of the left goal, new position is the
                        -- bottom-right corner of the left goal area
                        New_Ball_Position :=
                          Coordinates_3D_T'(Field.Left_Goal_Area.Right,
                                            Field.Left_Goal_Area.Bottom,
                                            0.0);
                     else
                        -- on the left of the left goal, new position is the
                        -- top-right corner of the left goal area
                        New_Ball_Position :=
                          Coordinates_3D_T'(Field.Left_Goal_Area.Right,
                                            Field.Left_Goal_Area.Top,
                                            0.0);
                     end if;
                  else
                     -- ball sent off on right side of the field
                     if Ball_Position.Y > Field.Bottom_Touchline / 2 then
                        -- on the left of the right goal, new position is the
                        -- bottom-left corner of the right goal area
                        New_Ball_Position :=
                          Coordinates_3D_T'(Field.Right_Goal_Area.Left,
                                            Field.Right_Goal_Area.Bottom,
                                            0.0);
                     else
                        -- on the right of the right goal, new position is the
                        -- top-left corner of the right goal area
                        New_Ball_Position :=
                          Coordinates_3D_T'(Field.Right_Goal_Area.Left,
                                            Field.Right_Goal_Area.Top,
                                            0.0);
                     end if;
                  end if;
                  -- create the goal kick event and push it in the buffer
                  Events_Buffer.Add(Goal_Kick_Event_T'(Team_Id => Offence_Team_Id));
                  -- set the new game state
                  New_Game_State := State_Record_T'(Game_Control.Goal_Kick,
                                                    Current_Status,
                                                    New_Ball_Position,
                                                    Offence_Team_Id);
                  Status.Teams_Status(New_Game_State.Offence_Team)
                    .Set_Starting_Player(New_Game_State);
                  Starting_Player := Identifier_T'
                    (Offence_Team_Id,
                     Status.Player_Mapper
                       (Offence_Team_Id,
                        Status.Teams_Status(Offence_Team_Id).Get_Starting_Player)
                     .Get_Player_Number);
                  Game_State.Set(New_Game_State);
                  -- wait for players
                  Game_State.Wait_For_Players;
                  -- create the restart event and push it into the buffer
                  Events_Buffer.Add(Restart_Event_T'(null record));
                  -- update the ball status; the new possessor is not set so it is
                  -- the last player which touch the ball (the new possessor will
                  -- be set with the id of the goalkeeper when he kicks the ball)
                  New_Ball_Motion := Motion.Get_Fixed_Motion(New_Ball_Position);
                  New_Ball_Status := (New_Ball_Motion, Starting_Player, False);
                  Status.Ball_Status.Set(New_Ball_Status);
                  -- set the new game state
                  delay until Status.Get_Corresponding_Quantum(Clock + Status.Quantum);
                  Game_State.Start;
               end;
            end if;
         or
            -- handle corner kick interruption
            accept Corner_Kick (Ball_Status : in Ball.Ball_Status_T) do
               -- the randevouz is used only for parameter passing
               Local_Ball_Status := Ball_Status;
            end Corner_Kick;
            -- create the interruption event and push it into the buffer
            Events_Buffer.Add(Interruption_Event_T'(Category => Ball_Out));
            declare
               -- local var to store the offence team id
               Offence_Team_Id : constant Team_Id_T :=
                 Player.Opposite_Team(Local_Ball_Status.Possessor.Team_Id);
               -- local var to store the (old) ball position
               Ball_Position : constant Coordinates_3D_T :=
                 Status.Ball_Status.Get.Motion.Get_Current_Position;
               -- local var to store the new ball position
               New_Ball_Position : Coordinates_3D_T;
               -- local var to store the new game state
               New_Game_State    : State_Record_T;
               -- local vars to store the new ball status
               New_Ball_Motion   : Motion.Motion_3d_T;
               New_Ball_Status : Ball.Ball_Status_T;
               Starting_Player : Identifier_T;
            begin
               -- handle substitutions (if any) for each team
               Handle_Substitutions(Output_Types.Local);
               Handle_Substitutions(Output_Types.Visitor);
               -- check where the ball has left the field to get the new ball
               -- position
               if Ball_Position.X < Field.Left_Goal_Line then
                  -- ball sent off on left side of the field
                  if Ball_Position.Y > Field.Bottom_Touchline / 2 then
                     -- on the right of the left goal, new position is the
                     -- bottom-left corner of the game field
                     New_Ball_Position :=
                       Coordinates_3D_T'(Field.Bottom_Left_Corner with 0.0);
                  else
                     -- on the left of the left goal, new position is the
                     -- top-left corner of the game field
                     New_Ball_Position :=
                       Coordinates_3D_T'(Field.Top_Left_Corner with 0.0);
                  end if;
               else
                  -- ball sent off on right side of the field
                  if Ball_Position.Y > Field.Bottom_Touchline / 2 then
                     -- on the left of the right goal, new position is the
                     -- bottom-right corner of the game field
                     New_Ball_Position :=
                       Coordinates_3D_T'(Field.Bottom_Right_Corner with 0.0);
                  else
                     -- on the right of the right goal, new position is the
                     -- top-right corner of the game field
                     New_Ball_Position :=
                       Coordinates_3D_T'(Field.Top_Right_Corner with 0.0);
                  end if;
               end if;
               -- create the korner kick event and push it into the buffer
               Events_Buffer.Add(Corner_Kick_Event_T'(Team_Id => Offence_Team_Id));
               -- set the new game state in which the
               New_Game_State := State_Record_T'(Game_Control.Corner_Kick,
                                                 Current_Status,
                                                 New_Ball_Position,
                                                 Offence_Team_Id);
               Game_State.Set(New_Game_State);
               -- wait for players
               Status.Teams_Status(Offence_Team_Id).Set_Starting_Player(New_Game_State);
               Starting_Player := Identifier_T'
                 (Offence_Team_Id,
                  Status.Player_Mapper
                    (Offence_Team_Id,
                     Status.Teams_Status(Offence_Team_Id).Get_Starting_Player)
                  .Get_Player_Number);
               Game_State.Wait_For_Players;
               -- create the restart event and push it into the buffer
               Events_Buffer.Add(Restart_Event_T'(null record));
               -- update the ball status; the new possessor is not set so it is
               -- the last player which touch the ball (the new possessor will
               -- be set when the goalkeeper kicks the ball)
               New_Ball_Motion := Motion.Get_Fixed_Motion(New_Ball_Position);
               Get_New_Ball_Status(New_Ball_Motion, Offence_Team_Id,
                                   New_Ball_Status);
               Status.Ball_Status.Set(New_Ball_Status);
               -- set the new game state
               delay until Status.Get_Corresponding_Quantum(Clock + Status.Quantum);
               Game_State.Start;
            end;
         or
            -- handle goal interruption
            accept Goal (Player_Id : in Identifier_T;
                         Team_Id : in Team_Id_T;
                         Is_Autogoal : in Boolean) do
               -- the randevouz is used only for parameter passing
               Local_Player_Id := Player_Id;
               Local_Team_Id := Team_Id;
               Local_Is_Autogoal := Is_Autogoal;
            end Goal;
            -- create the interruption event and push it into the buffer
            Events_Buffer.Add(Interruption_Event_T'(Category => Goal));
            -- handle substitutions (if any) for each team
            Handle_Substitutions(Output_Types.Local);
            Handle_Substitutions(Output_Types.Visitor);
            -- create and push the Goal_Event
            Events_Buffer.Add(Goal_Event_T'(Team_Id     => Local_Team_Id,
                                            Player_Id   => Local_Player_Id,
                                            Is_Autogoal => Local_Is_Autogoal));
            -- check if the current period is ended
            if Current_Period_Is_Ended then
               Handle_End_Period;
            else
               declare
                  -- local vars to store the new position (and motion) of the ball
                  -- when the game will restart, the id of the team which has to
                  -- restart the game and the new game state
                  -- local var to store the id of the team which has to restart game
                  Offence_Team_Id     : constant Team_Id_T :=
                    Player.Opposite_Team(Local_Team_Id);
                  -- local var to store the new ball position
                  New_Ball_Position   : constant Coordinates_3D_T :=
                    Coordinates_3D_T'
                      (Field.Midfield(Status.Teams_Status(Offence_Team_Id).Get_Side)
                       with 0.0);
                  -- local var to store the new game state
                  New_Game_State      : constant State_Record_T :=
                    State_Record_T'(Game_Control.Kick_Off,
                                    Current_Status,
                                    New_Ball_Position,
                                    Offence_Team_Id);
                  -- local vars to store the new ball status
                  New_Ball_Motion     : constant Motion.Motion_3d_T :=
                    Motion.Get_Fixed_Motion(New_Ball_Position);
                  Current_Ball_Status : constant Ball.Ball_Status_T :=
                    Status.Ball_Status.Get;
                  New_Ball_Status     : Ball.Ball_Status_T;
               begin
                  -- create the kick off event and push it to the buffer
                  Events_Buffer.Add(Kick_Off_Event_T'(Team_Id => Offence_Team_Id));
                  Status.Teams_Status(Offence_Team_Id)
                    .Set_Starting_Player(New_Game_State);
                  Get_New_Ball_Status(New_Ball_Motion, Offence_Team_Id,
                                      New_Ball_Status);
                  -- set the new game state
                  Game_State.Set(New_Game_State);
                  -- wait for players
                  Game_State.Wait_For_Players;
                  -- here all *active* players are ready to restart the game
                  -- create the restart event and push it into the buffer
                  Events_Buffer.Add(Restart_Event_T'(null record));
                  -- update the ball status
                  Status.Ball_Status.Set(New_Ball_Status);
                  -- set the new game state
                  delay until Status.Get_Corresponding_Quantum(Clock + Status.Quantum);
                  Game_State.Start;
               end;
            end if;
         or
            -- handle end half time break request
            accept End_Half_Time_Break  do
               -- this call is legal only during half time break
               if Current_Status = First_Half_Finished then
                  -- create the break time out event and push it into the buffer
                  Events_Buffer.Add(Break_Time_Out_Event_T'(Is_Whistle => False));
                  -- stop television timer
                  Television.Frame_Manager_Task.Stop_Timer;
                  -- start the second period
                  Start_Second_Period;
               else
                  -- we are not in the half time break, do nothing
                  null;
               end if;
            end End_Half_Time_Break;
         or
            -- handle substitution request
            accept Substitution (Substitution_Request 	: in Substitution_Request_T) do
               declare
                  -- local vars to store the incoming/outgoing player number
                  -- and the team id
                  Team_Id : constant Team_Id_T :=
                    Substitution_Request.Team_Id;
                  Incoming_Player_Number : constant Player_Number_T :=
                    Shirt_To_Player_Number(Team_Id).Element(Substitution_Request.Incoming_Shirt_Number);
                  Outgoing_Player_Number : constant Player_Number_T :=
                    Shirt_To_Player_Number(Team_Id).Element(Substitution_Request.Outgoing_Shirt_Number);
               begin
                  -- check if there are already too many pending request
                  if Substitutions_Registry(Team_Id).Length >=
                    Count_Type(Substitutions_Limit) then
                     -- already too much pending requests, deny
                     -- create the substitution request event and push it to the
                     -- buffer
                     Events_Buffer.Add(
                       Substitution_Request_Event_T'(
                         Outgoing_Player_Id => (Team_Id,Outgoing_Player_Number),
                         Is_Accepted        => False));
                  else
                     -- ok, allow substitution request
                     if Current_Status = First_Half_Finished then
                        -- we are during half-time break so we can handle
                        -- substitution immediately, check substitution admissibility
                        -- TODO: this code is repeated, move into a local function
                        if Check_Admissibility(Substitution_Request) then
                           -- create the substitution event and push it into the buffer
                           Events_Buffer
                             .Add(Substitution_Event_T'(
                               Incoming_Player_Id => (Team_Id,Incoming_Player_Number),
                               Outgoing_Player_Id => (Team_Id,Outgoing_Player_Number),
                               Agreed             => True));
                           -- substitute player and increment the number of
                           -- substitution made by team
                           Status.Players_Status(Team_Id,Outgoing_Player_Number)
                             .Substitute(Incoming_Player_Number);
                           Substitutions_Num(Team_Id) := Substitutions_Num(Team_Id) + 1;
                        else
                           -- substitution in not legal, deny it
                           -- create the substitution event and push it into the buffer
                           Events_Buffer
                             .Add(Substitution_Event_T'(
                               Incoming_Player_Id => (Team_Id,Incoming_Player_Number),
                               Outgoing_Player_Id => (Team_Id,Outgoing_Player_Number),
                               Agreed             => False));
                        end if;
                     else
                        -- game is in any other state (play), check whether the
                        -- substitution is currently admissible and eventually
                        -- register the substitution request
                        if Check_Admissibility(Substitution_Request) then
                           -- create the substitution request event and push it
                           -- to the buffer
                           Events_Buffer
                             .Add(Substitution_Request_Event_T'(
                               Outgoing_Player_Id => (Team_Id,Outgoing_Player_Number),
                               Is_Accepted        => True));
                           Substitutions_Registry(Team_Id).Append(Substitution_Request);
                        else
                           -- substitution is not legal, denay it
                           -- create the substitution request event and push it
                           -- to the buffer
                           Events_Buffer
                             .Add(Substitution_Request_Event_T'(
                               Outgoing_Player_Id => (Team_Id,Outgoing_Player_Number),
                               Is_Accepted        => False));
                        end if;
                     end if;
                  end if;
               end;
            end Substitution;
         or
            accept Abort_Match;
            Current_Status := Match_Aborted;
            -- create the state change event and push it into the buffer
            Events_Buffer.Add(State_Change_Event_T'(State => Current_Status));
            -- create the abort event and push it into the buffer
            Events_Buffer.Add(Abort_Match_Event_T'(null record));
         or
            when not Current_Period_Is_Ended =>
               -- in the meantime wait for the end of current period or the end of
               -- the half time break
               delay until Time_Out;
               -- we are here if the current period or the halt fime break ended
               if Current_Status = First_Half_Finished then
                  -- create the break time out event and push it into the buffer
                  Events_Buffer.Add(Break_Time_Out_Event_T'(Is_Whistle => False));
                  -- half time break ended, restart game with second period
                  Start_Second_Period;
               else
                  -- current period is ended, register the end
                  Current_Period_Is_Ended := True;
                  if Current_Status = Match_Started then
                     -- first period ended, create the first half time out event
                     -- (without whistle) and push it into the buffer
                     Events_Buffer
                       .Add(First_Half_Time_Out_Event_T'(Is_Whistle => False));
                  elsif Current_Status = Second_Half_Started then
                     -- second period ended, create the second half time out event
                     -- (without whistle) and push it into the buffer
                     Events_Buffer
                       .Add(Second_Half_Time_Out_Event_T'(Is_Whistle => False));
                  end if;
               end if;
         end select;
      end loop;
      -- finalization
      Game_State.Wait_The_End;
      -- stop the assistant referee
      Assistant_Referee.Assistant_Referee_Task.Quit;
      -- stop the television task
      Television.Frame_Manager_Task.Quit;
      -- stop the Events_Manager_Task
      Events_Buffer.Add(Termination_Event_T'(null record));
   end Referee_Task;
end Referee;
