with Position; use Position;

Package body Tactic is
   function Outside_Position(Side: Side_T; Role_Index: Player_Index_T)
                             return Coordinates_T  is
   begin
      if Side = Left then
         return Coordinates_T'(Left_Midfield.Right - 12 + Role_Index,
                               Bottom_Touchline + 2);
      else
         return Coordinates_T'(Right_Midfield.Left + 12 - Role_Index,
                               Bottom_Touchline + 2);
      end if;
   end Outside_Position;

   -- Get_Position returns the Kick-off position for a given player
   -- TODO: Improve with the Game_State information when Game_State
   -- will be available. := 0
   function Get_Standard_Position (Formation	: Formation_T;
                                   Side		: Side_T;
                                   Role_Index	: Player_Index_T)
                                   return Coordinates_T
   is
      Standard_Position : Coordinates_T;
   begin
      if Side = Left then
         Standard_Position :=
           (Formation.Player_Positions(Role_Index).X,
            Formation.Player_Positions(Role_Index).Y);
      else
         Standard_Position :=
           (Pitch.Right + 1
            - Formation.Player_Positions(Role_Index).X,
            Pitch.Bottom + 1
            - Formation.Player_Positions(Role_Index).Y);
      end if;
      return Standard_Position;
   end Get_Standard_Position;

   function Right_Bound
     (Formation: Formation_T; Side: Side_T; Player: Player_Index_T)
      return X_Range_T is
   begin
      if Side = Left then
         return Formation.Player_Zones(Player).Right;
      else
         return Right_Goal_Line - Formation.Player_Zones(Player).Left;
      end if;
   end;

   function Left_Bound
     (Formation: Formation_T; Side: Side_T; Player: Player_Index_T)
      return X_Range_T is
   begin
      if Side = Left then
         return Formation.Player_Zones(Player).Left;
      else
         return Right_Goal_Line - Formation.Player_Zones(Player).Right;
      end if;
   end;

   function Upper_Bound
     (Formation: Formation_T; Side: Side_T; Player: Player_Index_T)
      return Y_Range_T is
   begin
      if Side = Left then
         return Formation.Player_Zones(Player).Top;
      else
         return Bottom_Touchline - Formation.Player_Zones(Player).Bottom;
      end if;
   end;

   function Lower_Bound
     (Formation: Formation_T; Side: Side_T; Player: Player_Index_T)
      return Y_Range_T is
   begin
      if Side = Left then
         return Formation.Player_Zones(Player).Bottom;
      else
         return Bottom_Touchline - Formation.Player_Zones(Player).Top;
      end if;
   end;

   function Ball_Proportional_Position(Formation: 	Formation_T;
                                       Side: 		Side_T;
                                       Role_Index:	Player_Index_T;
                                       Ball_Position:	in Coordinates_3D_T;
                                       Pressing:	Integer)
                                       return Coordinates_T
   is
      Position	: Coordinates_T;	-- The function result
      Ball_X_Delta : constant Integer:= Ball_Position.X - Midfield_Line;
      Ball_Y_Delta : constant Integer:= Ball_Position.Y - Bottom_Touchline / 2;
      Player_Max_Delta_X : Integer;
      Player_Max_Delta_Y : Integer;
   begin
      Position := Get_Standard_Position(Formation, Side, Role_Index);
      -- Adapt position with pressing
      if (Pressing > 0 and Side = Left) then
         Position.X := Position.X +
           (Right_Bound(Formation, Side, Role_Index) - Position.X) * Pressing / 5;
      elsif (Pressing < 0 and Side = Left) then
         Position.X := Position.X +
           (Position.X - Left_Bound(Formation, Side, Role_Index)) * Pressing / 5;
      elsif (Pressing > 0 and Side = Right) then
         Position.X := Position.X -
           (Position.X - Left_Bound(Formation, Side, Role_Index)) * Pressing / 5;
      elsif (Pressing < 0 and Side = Right) then
         Position.X := Position.X -
           (Right_Bound(Formation, Side, Role_Index) - Position.X) * Pressing / 5;
      end if;
      -- Calculate the max displacement for the player along
      -- the X axis...
      if (Ball_Position.X > Pitch.Right / 2) then
         Player_Max_Delta_X := Right_Bound(Formation, Side, Role_Index) - Position.X;
      else
         Player_Max_Delta_X := Position.X - Left_Bound(Formation, Side, Role_Index);
      end if;
      -- ...and the Y axis
      if (Ball_Position.Y >= Bottom_Touchline / 2) then
         Player_Max_Delta_Y := Lower_Bound(Formation, Side, Role_Index)
           - Position.Y;
      else
         Player_Max_Delta_Y := Position.Y - Upper_Bound(Formation, Side, Role_Index);
      end if;
      -- Calculate the correct player Position

      Position.X := Position.X
        + Player_Max_Delta_X * Ball_X_Delta / Midfield_Line; -- ball
      Position.Y := Position.Y
        + Integer(Float(Player_Max_Delta_Y * Ball_Y_Delta) * 2.0
                  / Float(Bottom_Touchline));
      return Position;
   end;

   function Reference_Position (Formation: 	Formation_T;
                                Side: 		Side_T;
                                Role_Index: 	Player_Index_T;
                                State: 		State_T;
                                Ball_Position:	in Coordinates_3D_T;
                                Pressing:	Integer)
                                return Coordinates_T
   is
      Position	: Coordinates_T;	-- The function result
   begin
      case State is
         -- NOTE: for game interruptions, the player who restarts the game
         -- should *NOT* follow this scheme.  He should just reach the position
         -- that will be assigned to the ball by the referee.
         when Play =>
            Position :=
              Ball_Proportional_Position(Formation, Side, Role_Index, Ball_Position, Pressing);
         when Kick_Off =>
            Position := Get_Standard_Position(Formation, Side, Role_Index);
            -- Change the player position to send him on his own field side
            if (Role_Index > 1 + Formation.Players_By_Role(Defender)) then
               if Side = Left then
                  Position.X := Integer(0.6 * Float(Position.X));
               else
                  Position.X := Right_Goal_Line + 1
                    - Integer(0.6 * Float(Pitch.Right + 1 - Position.X));
               end if;
            end if;
         when Throw_In =>
            Position :=
              Ball_Proportional_Position(Formation, Side, Role_Index, Ball_Position, Pressing);
         when Goal_Kick =>
            Position :=
              Ball_Proportional_Position(Formation, Side, Role_Index, Ball_Position, Pressing);
         when Corner_Kick =>
            Position :=
              Ball_Proportional_Position(Formation, Side, Role_Index, Ball_Position, Pressing);
            -- The defending goalkeeper stays on the mid of the goal
            -- PRE: the ball is expected to be on (or beyond) the goal line!
            if -- the corner kick is played on the LEFT side and the goalkeeper
            -- is the one playing on *that* side
              (Ball_Position.X <= Left_Goal_Line and Role_Index = 1 and Side = Left)
              or -- the corner kick is played on the RIGHT side and
            -- the goalkeeper is the one playing on *that* side
              (Ball_Position.X >= Right_Goal_Line and Role_Index = 1 and Side = Right)
            then
               Position.Y := Bottom_Touchline / 2;
            end if;
         when Free_Kick =>
            Position :=
              Ball_Proportional_Position(Formation, Side, Role_Index, Ball_Position, Pressing);
            -- TODO: Players should stay 9 m away from the ball...
         when Penalty_Kick =>
            Position :=
              Ball_Proportional_Position(Formation, Side, Role_Index, Ball_Position, Pressing);
            -- TODO: Players should stay 9 m away behind the ball... does it
            -- still make sense to call this function for a Penalty kick?
         when Half_Time_Break..End_Match =>
            Position := Outside_Position(Side, Role_Index);
         when Not_Started => -- Used at the beginning
            Position := Outside_Position(Side, Role_Index);
            -- Position := Get_Standard_Position(Formation, Side, Role_Index);
      end case;
      return Position;
   end Reference_Position;
end Tactic;
