with Field; use Field;
with Game_Control; use Game_Control;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Output_Types; use Output_Types;

-- describes the most common soccer formations such as 4-4-2, 4-3-3, etc.
--  *	the player 'standard' position for each formation
--  *	the zones in which the players move
--  *	methods to calculate a 'reference' position according to the
--	current game state

package Tactic is
   -- the role associated to players
   type Player_Role_T is (Goalkeeper, Defender, Midfielder, Forward);
   type Player_Standard_Positions_T is array (1..11) of Coordinates_T;
   type Player_Zones_T is array (1..11) of Field_Area_T;
   type Players_By_Role_T is array (Player_Role_T range Defender..Forward)
     of Player_Index_T;

   -- i.e. 4-4-2 means "4 defenders, 4 midfielders, 4 Forwards"
   type Formation_T is record
      Id : Natural;
      Name : Unbounded_String;
      Players_By_Role : Players_By_Role_T;
      Player_Positions : Player_Standard_Positions_T;
      Player_Zones: Player_Zones_T;
   end record;

   function Outside_Position(Side: Side_T; Role_Index: Player_Index_T)
                             return Coordinates_T;

   function Get_Standard_Position(Formation: Formation_T;
                                  Side: Side_T;
                                  Role_Index: Player_Index_T) return Coordinates_T;

   function Reference_Position (Formation : Formation_T;
                                Side : Side_T;
                                Role_Index : Player_Index_T;
                                State : State_T;
                                Ball_Position : Coordinates_3D_T;
                                Pressing : Integer)
                                return Coordinates_T;
   -- <summary>Calculate a reference position according to the game state.
   -- </summary>
   -- <description>
   -- Each player has an associated zone, defined according to his role.
   -- When they don't have the ball, players stay inside their own zone
   -- and they move back and forth proportionally to the position of the ball
   -- inside the field.
   -- IMPORTANT: This function
   -- * is *NOT* supposed to return the position for
   --	the player(s) who should restart the game after an interruption
   -- 	if that position depends somehow on the team tactics (i.e. no check
   -- 	is made to see who is expected to throw the ball first);
   -- * is not actually checking whether the calculated position is coherent
   --	with the rules of the game (i.e. if the player and the ball are
   --	distant enough.
   -- </description>
   -- <parameter name="Formation">The current formation (4-4-2, ...)</parameter>
   -- <parameter name="Side">The side of the player's own goal</parameter>
   -- <parameter name="Role">Maps the player to his std pos</parameter>
   -- <parameter name="State">The current game state</parameter>
   -- <parameter name="Ball_Position">The current ball position when State is
   -- 'Play', otherwise the position that will be assigned to the ball
   -- by the referee.
   -- </parameter>

   P_442 : Player_Standard_Positions_T :=
     ((6, 35), (24, 7), (18, 27), (18, 42), (24, 62), -- goalkeeper & defenders
      (60, 7), (52, 25), (52, 43), (60, 62),    -- midfielders
      (86, 25), (86,44));                       -- forwards

   Z_442 : Player_Zones_T :=
   -- pos 1 ) Goalkeeper
     ((Left => 1, Right => 17, Top => Goal_Y_Min, Bottom => Goal_Y_Max),
      -- DEFENDERS (backward 50%, forward up to midfield, left/right 33%)
      -- pos 2) left
      (Left => P_442(2).X / 2, Right => Midfield_Line,
       Top => P_442(2).Y - (P_442(2).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(2).Y + (Bottom_Touchline - P_442(2).Y) * 1 / 3),
      -- pos 3) center left fullback
      (Left => P_442(3).X / 2, Right => Left_Midfield.Right,
       Top => P_442(3).Y - (P_442(3).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(3).Y + (Bottom_Touchline - P_442(3).Y) * 1 / 3),
      -- pos 4) center right fullback
      (Left => P_442(4).X / 2, Right => Left_Midfield.Right,
       Top => P_442(4).Y - (P_442(4).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(4).Y + (Bottom_Touchline - P_442(4).Y) * 1 / 3),
      -- pos 5) right fullback
      (Left => P_442(5).X / 2, Right => Left_Midfield.Right,
       Top => P_442(5).Y - (P_442(5).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(5).Y + (Bottom_Touchline - P_442(5).Y) * 1 / 3),
      -- pos 6-9) midfielders
      (Left => P_442(6).X / 2, Right => Left_Midfield.Right * 3 / 2,
       Top => P_442(6).Y - (P_442(6).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(6).Y + (Bottom_Touchline - P_442(6).Y) * 1 / 3),
      (Left => P_442(7).X / 2, Right => Left_Midfield.Right * 3 / 2,
       Top => P_442(7).Y - (P_442(7).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(7).Y + (Bottom_Touchline - P_442(7).Y) * 1 / 3),
      (Left => P_442(8).X / 2, Right => Left_Midfield.Right * 3 / 2,
       Top => P_442(8).Y - (P_442(8).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(8).Y + (Bottom_Touchline - P_442(8).Y) * 1 / 3),
      (Left => P_442(9).X / 2, Right => Left_Midfield.Right * 3 / 2,
       Top => P_442(9).Y - (P_442(9).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(9).Y + (Bottom_Touchline - P_442(9).Y) * 1 / 3),
      -- Forwards
      (Left => P_442(10).X / 2,
       Right => P_442(10).X + ((Right_Goal_Line - P_442(10).X)/2),
       Top => P_442(10).Y - (P_442(10).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(10).Y + (Bottom_Touchline - P_442(10).Y) * 1 / 3),
      (Left => P_442(11).X / 2,
       Right => P_442(10).X + ((Right_Goal_Line - P_442(10).X)/2),
       Top => P_442(11).Y - (P_442(11).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_442(11).Y + (Bottom_Touchline - P_442(11).Y) * 1 / 3)
     ); -- TODO: define the zone of each player

   F_442 : constant Formation_T :=
     (Id => 1,
      Name => To_Unbounded_String("4-4-2"),
      Players_By_Role => (Defender => 4, Midfielder => 4, Forward => 2),
      Player_Positions => P_442,
      Player_Zones     => Z_442);

   P_433 : Player_Standard_Positions_T :=
     ((6, 35), (24, 7), (18, 27), (18, 42), (24, 62), -- goalkeeper & defenders
      (56, 17), (52, 34), (56, 51),    -- midfielders
      (86, 17), (76, 34), (86,51));  -- forwards


   Z_433 : Player_Zones_T :=
   -- pos 1 ) Goalkeeper
     ((Left => 1, Right => 17, Top => Goal_Y_Min, Bottom => Goal_Y_Max),
      -- DEFENDERS (backward 50%, forward up to midfield, left/right 33%)
      -- pos 2) left
      (Left => P_433(2).X / 2, Right => Midfield_Line,
       Top => P_433(2).Y - (P_433(2).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(2).Y + (Bottom_Touchline - P_433(2).Y) * 1 / 3),
      -- pos 3) center left fullback
      (Left => P_433(3).X / 2, Right => Left_Midfield.Right,
       Top => P_433(3).Y - (P_433(3).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(3).Y + (Bottom_Touchline - P_433(3).Y) * 1 / 3),
      -- pos 4) center right fullback
      (Left => P_433(4).X / 2, Right => Left_Midfield.Right,
       Top => P_433(4).Y - (P_433(4).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(4).Y + (Bottom_Touchline - P_433(4).Y) * 1 / 3),
      -- pos 5) right fullback
      (Left => P_433(5).X / 2, Right => Left_Midfield.Right,
       Top => P_433(5).Y - (P_433(5).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(5).Y + (Bottom_Touchline - P_433(5).Y) * 1 / 3),
      -- pos 6-8) midfielders
      (Left => P_433(6).X / 2, Right => Left_Midfield.Right * 3 / 2,
       Top => P_433(6).Y - (P_433(6).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(6).Y + (Bottom_Touchline - P_433(6).Y) * 1 / 3),
      (Left => P_433(7).X / 2, Right => Left_Midfield.Right * 3 / 2,
       Top => P_433(7).Y - (P_433(7).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(7).Y + (Bottom_Touchline - P_433(7).Y) * 1 / 3),
      (Left => P_433(8).X / 2, Right => Left_Midfield.Right * 3 / 2,
       Top => P_433(8).Y - (P_433(8).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(8).Y + (Bottom_Touchline - P_433(8).Y) * 1 / 3),
      --forwards
      (Left => P_433(9).X / 2, Right => Left_Midfield.Right * 3 / 2,
       Top => P_433(9).Y - (P_433(9).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(9).Y + (Bottom_Touchline - P_433(9).Y) * 1 / 3),
      (Left => P_433(10).X / 2,
       Right => P_433(10).X + ((Right_Goal_Line - P_433(10).X)/2),
       Top => P_433(10).Y - (P_433(10).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(10).Y + (Bottom_Touchline - P_433(10).Y) * 1 / 3),
      (Left => P_433(11).X / 2,
       Right => P_433(10).X + ((Right_Goal_Line - P_433(10).X)/2),
       Top => P_433(11).Y - (P_433(11).Y - Top_Left_Corner.Y) * 1/3,
       Bottom => P_433(11).Y + (Bottom_Touchline - P_433(11).Y) * 1 / 3)
     ); -- TODO: define the zone of each player

   F_433 : constant Formation_T :=
     (Id => 2,
      Name => To_Unbounded_String("4-3-3"),
      Players_By_Role => (Defender => 4, Midfielder => 3, Forward => 3),
      Player_Positions => P_433,
      Player_Zones     => Z_433);

   Formations_Array : constant array (1..2) of Formation_T := (F_442, F_433);
end Tactic;
