with Events; use Events;
with Schema.Validators; use Schema.Validators;
with Input_Sources.File; use Input_Sources.File;
with Schema.Schema_Readers; use Schema.Schema_Readers;
with Schema.Dom_Readers; use Schema.Dom_Readers;
with String_Stream; use String_Stream;
with Sax.Readers; use Sax.Readers;
with DOM.Core.Documents; use Dom.Core.Documents;
with DOM.Core.Nodes;	 use DOM.Core.Nodes;
with DOM.Core.Attrs;	 use DOM.Core.Attrs;
with Ada.Text_IO; use Ada.Text_IO;
with Status; use Status;
with Field; use Field;
with Tactic; use Tactic;
with Ada.Strings; use Ada.Strings;
with DOM.Core.Elements; use DOM.Core.Elements;

package body Configuration is
   -- XML document tree for simulation configuration
   Simulation_Config_Xml_Doc : Document;
   -- XML document tree for teams configuration
   Teams_Config_Xml_Doc : Document;

   -- a private procedure to validate and parse a XML document with an XML schema;
   -- the procedure returns the xml tree parsed and its the string representation
   procedure Validate_And_Parse_Xml_File(Xml_File_Path		: in String;
                                         Schema_File_Path	: in String;
                                         Xml_Tree		: out Document;
                                         Xml_Tree_String	: out Unbounded_String) is
      -- the grammar defined by the schema
      Schema_Grammar : XML_Grammar;
      -- the input file used to read the schema file
      Schema_Input : File_Input;
      -- the schema reader
      Schema_Reader : Schema.Schema_Readers.Schema_Reader;
      -- the input file user to read the xml file
      Xml_Input : File_Input;
      -- the xml reader
      Xml_Tree_Reader : Schema.Dom_Readers.Tree_Reader;
      -- the string stream used to create the string representation of the xml tree
      Tree_String_Output : aliased String_Stream_Type;
   begin
      -- open and parse the schema file
      Set_Public_Id (Xml_Input, "XML schema file");
      Open(Schema_File_Path, Schema_Input);
      Parse(Schema_Reader, Schema_Input);
      Close(Schema_Input);
      -- get the schema grammar and set it to the xml reader
      Schema_Grammar := Get_Grammar(Schema_Reader);
      Set_Grammar(Xml_Tree_Reader, Schema_Grammar);
      -- open xml file
      Set_Public_Id (Xml_Input, "XML file");
      Open(Xml_File_Path, Xml_Input);
      -- set parsing features
      Set_Feature (Xml_Tree_Reader, Validation_Feature, False);
      Set_Feature (Xml_Tree_Reader, Schema_Validation_Feature, True);
      Set_Feature (Xml_Tree_Reader, Namespace_Feature, False);
      -- parse xml file
      Parse(Xml_Tree_Reader, Xml_Input);
      -- close xml file
      Close(Xml_Input);
      -- get the xml tree and its string representation
      Xml_Tree := Get_Tree(Xml_Tree_Reader);
      Open(Tree_String_Output, "");
      Write(Tree_String_Output'Access, Xml_Tree);
      Xml_Tree_String := Tree_String_Output.Str;
   exception
      when XML_Validation_Error =>
         Put_Line("XML VALIDATION ERROR: " & Get_Error_Message(Xml_Tree_Reader));
   end Validate_And_Parse_Xml_File;

   procedure Read_Simulation_Configuration(Simulation_Xml_String : out Unbounded_String) is
   begin
      -- validate and parse simulation configuration file
      Validate_And_Parse_Xml_File("config/simulation.xml",
                                  "config/simulation_schema.xsd",
                                  Simulation_Config_Xml_Doc,
                                  Simulation_Xml_String);
   end Read_Simulation_Configuration;

   procedure Read_Teams_Configuration(Teams_Xml_String : out Unbounded_String) is
   begin
      -- validate and parse teams configuration file
      Validate_And_Parse_Xml_File("config/teams.xml",
                                  "config/teams_schema.xsd",
                                  Teams_Config_Xml_Doc,
                                  Teams_Xml_String);
   end Read_Teams_Configuration;

   procedure Load_Default_Simulation(Configuration : out Simulation_Configuration_T) is
      Period_Duration_Node : Node;
      Break_Duration_Node : Node;
      Substitutions_Allowed_Node : Node;
   begin
      -- get simulation configuration default values
      Period_Duration_Node :=
        Item(DOM.Core.Documents.Get_Elements_By_Tag_Name(Simulation_Config_Xml_Doc, "periodDuration"),0);
      Configuration.Period_Duration :=
        Natural'Value(Value(Get_Named_Item(Attributes(Period_Duration_Node), "default")));
      Break_Duration_Node :=
        Item(DOM.Core.Documents.Get_Elements_By_Tag_Name(Simulation_Config_Xml_Doc, "breakDuration"),0);
      Configuration.Break_Duration :=
        Natural'Value(Value(Get_Named_Item(Attributes(Break_Duration_Node), "default")));
      Substitutions_Allowed_Node :=
        Item(DOM.Core.Documents.Get_Elements_By_Tag_Name(Simulation_Config_Xml_Doc, "substitutionsAllowed"),0);
      Configuration.Substitutions_Allowed :=
        Natural'Value(Value(Get_Named_Item(Attributes(Substitutions_Allowed_Node), "default")));
   end Load_Default_Simulation;

   procedure Set_Player_Characteristics(Team_Id : in Team_Id_T;
                                        Player_Number : in Natural;
                                        Current_Player_Node : in Node) is
   begin
      Players_Characteristics(Team_Id,Player_Number).Name :=
        To_Unbounded_String(
                            Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "name"),0))));
      Players_Characteristics(Team_Id,Player_Number).Shirt_Number :=
        Positive'Value(
                       Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "shirtNumber"),0))));

      Status.Shirt_To_Player_Number(Team_Id).Insert
        (Players_Characteristics(Team_Id,Player_Number).Shirt_Number, Player_Number);

      Players_Characteristics(Team_Id,Player_Number).Speed :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "speed"),0))));
      Players_Characteristics(Team_Id,Player_Number).Stamina :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "stamina"),0))));
      declare
         Passing_Element : constant Node := Item(
                                                 DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "passing"),0);
         Shooting_Element : constant Node := Item(
                                                  DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "shooting"),0);
         Heading_Element : constant Node := Item(
                                                 DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "heading"),0);
      begin
         Players_Characteristics(Team_Id,Player_Number).Passing_Power :=
           Float'Value(
                       Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Passing_Element, "power"),0))));
         Players_Characteristics(Team_Id,Player_Number).Passing_Accuracy :=
           Float'Value(
                       Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Passing_Element, "accuracy"),0))));
         Players_Characteristics(Team_Id,Player_Number).Shooting_Power :=
           Float'Value(
                       Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Shooting_Element, "power"),0))));
         Players_Characteristics(Team_Id,Player_Number).Shooting_Accuracy :=
           Float'Value(
                       Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Shooting_Element, "accuracy"),0))));
         Players_Characteristics(Team_Id,Player_Number).Heading_Power :=
           Float'Value(
                       Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Heading_Element, "power"),0))));
         Players_Characteristics(Team_Id,Player_Number).Heading_Accuracy :=
           Float'Value(
                       Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Heading_Element, "accuracy"),0))));
      end;
      Players_Characteristics(Team_Id,Player_Number).Height :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "height"),0))));
      Players_Characteristics(Team_Id,Player_Number).Jump_Skill :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "jumpSkill"),0))));
      Players_Characteristics(Team_Id,Player_Number).Dribble_Accuracy :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "dribbleAccuracy"),0))));
      Players_Characteristics(Team_Id,Player_Number).Tackle_Skill :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "tackleSkill"),0))));
      Players_Characteristics(Team_Id,Player_Number).Catching_Accuracy :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "catchingAccuracy"),0))));
      Players_Characteristics(Team_Id,Player_Number).Aggression :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "aggression"),0))));
      Players_Characteristics(Team_Id,Player_Number).Teak_Work_Skill :=
        Float'Value(
                    Node_Value(First_Child(Item(DOM.Core.Elements.Get_Elements_By_Tag_Name(Current_Player_Node, "teamWorkSkill"),0))));
   end Set_Player_Characteristics;

   procedure Load_Players(Team_Id		: in Team_Id_T;
                          Players_Configuration : in Distribution_Control.Players_Configuration_T) is
      -- a positive used to keep player tasks identifier
      Task_Id : Positive := 1;
   begin
      -- iterate on players
      for I in Players_Configuration'Range loop
         declare
            Current_Player_Node : constant Node :=
              Get_Element_By_Id(Teams_Config_Xml_Doc,
                                To_String(Players_Configuration(I).Player_Id));
         begin
            -- set all player characteristics
            Set_Player_Characteristics(Team_Id, I, Current_Player_Node);
            -- check if current player is first-string
            if Players_Configuration(I).Is_First_String then
               -- set player state and role
               Players_Status(Team_Id,I).Set_State(Active);
               Players_Status(Team_Id,I).Set_Role(
                                                  Players_Configuration(I).Player_Role_Index);
               -- set player player number and position
               Player_Mapper(Team_Id,Task_Id).Set_Numbers
                 (I, Players_Characteristics(Team_Id, I).Shirt_Number);
               Player_Mapper(Team_Id,Task_Id).Set_Position(
                                                           Outside_Position(Teams_Status(Team_Id).Get_Side, Task_Id));

               Task_Id := Task_Id + 1;
            end if;
         end;
      end loop;
   end Load_Players;

   procedure Load_Default_Players(Team_Id : in Team_Id_T;
                                  Xml_Team_Id : in Unbounded_String) is
      -- the team xml DOM node
      Team_Node : constant Node := Get_Element_By_Id(Teams_Config_Xml_Doc,
                                                     To_String(Xml_Team_Id));
      -- the first player xml DOM node
      Current_Player_Node : Node := First_Child(Team_Node);
      -- a natural used to keep players number
      Player_Number : Natural := 0;
      -- a positive used to keep player tasks identifier
      Task_Id : Positive := 1;
      I : Positive := 1;
   begin
      -- iterate on team players
      while Current_Player_Node /= null and I <= 17 loop
         -- process only element node
         if Node_Type(Current_Player_Node) = Element_Node then
            declare
               -- a costant to store current player default state
               Default_State : constant String :=
                 Value(Get_Named_Item(Attributes(Current_Player_Node), "defaultState"));
            begin
               Player_Number := Player_Number + 1;
               -- process only candidate players
               if Default_State /= "out" then
                  -- set all the static characteristics
                  Set_Player_Characteristics(Team_Id,
                                             Player_Number,
                                             Current_Player_Node);
                  -- set configuration for first-string players
                  if Default_State = "first-string" then
                     -- set player state and role
                     Players_Status(Team_Id,Player_Number).Set_State(Active);
                     Players_Status(Team_Id,Player_Number).Set_Role(
                       Player_Index_T'Value(
                         Value(Get_Named_Item(Attributes(Current_Player_Node), "defaultRole"))));
                     -- set player player number and position
                     Player_Mapper(Team_Id,Task_Id).Set_Numbers
                       (Player_Number,
                        Players_Characteristics(Team_Id, Player_Number).Shirt_Number);
                     Player_Mapper(Team_Id,Task_Id).Set_Position(
                                                                 Outside_Position(Teams_Status(Team_Id).Get_Side, Task_Id));

                     Task_Id := Task_Id + 1;
                  end if;
               end if;
            end;
            I := I + 1;
         end if;
         -- load the next player
         Current_Player_Node := Next_Sibling(Current_Player_Node);
      end loop;
   end Load_Default_Players;

   procedure Load_Team(Team_Id			: in Team_Id_T;
                       Xml_Team_Id		: in Unbounded_String;
                       Teams_Configuration 	: in Distribution_Control.Teams_Configuration_T;
                       Use_Default 		: in Boolean) is
      -- xml team DOM node
      Team_Node        : constant Node :=
        Get_Element_By_Id(Teams_Config_Xml_Doc, To_String(Xml_Team_Id));
      -- team formation and pressing value
      Formation_Id     : Natural;
      Pressing_Balance : Integer;

      Team_Name : constant String := Value(Get_Named_Item(Attributes(Team_Node), "name"));
   begin
      -- set the team id
      Teams_Status(Team_Id).Set_Id(Xml_Team_Id);
      -- set the team name
      Teams_Status(Team_Id)
        .Set_Name(To_Unbounded_String(
          Value(Get_Named_Item(Attributes(Team_Node), "name"))));
      Statistics.Teams_Statistics(Team_Id).Name(Team_Name'First..Team_Name'Last) :=
        Team_Name;
      -- set team side (we assign always left side to team A and right side to
      -- team B
      if Team_Id = Local then
         Teams_Status(Team_Id).Set_Side(Left);
      else
         Teams_Status(Team_Id).Set_Side(Right);
      end if;
      -- check if we have to load default configuration
      if Use_Default then
         -- get default formation and default pressing balance from XML
         Formation_Id := Natural'Value(
                                        Value(Get_Named_Item(Attributes(Team_Node), "defaultFormation")));
         Pressing_Balance := Integer'Value(
                                           Value(Get_Named_Item(Attributes(Team_Node), "defaultPressing")));
         -- load players default configuration
         Load_Default_Players(Team_Id, Xml_Team_Id);
         -- set the team configuration in the teams configuration protected object;
         -- this is needed for distribution control
         Distribution_Control.Teams_Configuration.Set_Default(Team_Id        => Team_Id,
                                                              Xml_Team_Id    => Xml_Team_Id,
                                                              Formation_Id   => Formation_Id,
                                                              Pressing_Value => Pressing_Balance);
      else
         -- get formation and pressing balance value from team configuration
         Formation_Id := Teams_Configuration(Team_Id).Formation_Id;
         Pressing_Balance := Teams_Configuration(Team_Id).Pressing_Balance_Value;
         -- load players configuration
         -- TODO: should be Load_Players
         Load_Players(Team_Id,
                      Teams_Configuration(Team_Id).Players_Configuration);
      end if;
      -- set team formation and pressing balance
      --Teams_Status(Team_Id).Set_Formation(Tactic.Formations_Array(Formation_Id));
      --Teams_Status(Team_Id).Set_Pressing(Pressing_Balance);
      Teams_Status(Team_Id).Set_Tactic(Tactic.Formations_Array(Formation_Id),
                                       Pressing_Balance);
   end Load_Team;

   procedure Load_Teams_Configuration is
      -- an array of two elements to store teams configuration
      Teams_Configuration : Distribution_Control.Teams_Configuration_T;
      -- teams xml identifiers
      Local_Team_Id : Unbounded_String;
      Visitor_Team_Id : Unbounded_String;
   begin
      -- get teams configuration from "distribution"
      Teams_Configuration := Distribution_Control.Teams_Configuration.Get;
      Local_Team_Id := Teams_Configuration(Local).Team_Id;
      Visitor_Team_Id := Teams_Configuration(Visitor).Team_Id;
      -- check if there is the configuration for Local team and if so get it
      if Local_Team_Id /= Null_Unbounded_String then
         Load_Team(Local, Local_Team_Id, Teams_Configuration, False);
      end if;
      -- check if there is the configuration for Visitor team and if so get it
      if Visitor_Team_Id /= Null_Unbounded_String then
         Load_Team(Visitor, Visitor_Team_Id, Teams_Configuration, False);
      end if;
      -- if there is not the configuration for Local get the default configuration
      -- that do not conflict with the Visitor configuration (if any)
      if Teams_Status(Local).Get_Id = Null_Unbounded_String then
         declare
            Current_Element : Node :=
              First_Child(DOM.Core.Documents.Get_Element(Teams_Config_Xml_Doc));
         begin
            -- get the first team xml element
            while Node_Type(Current_Element) /= Element_Node loop
               Current_Element := Next_Sibling(Current_Element);
            end loop;
            Local_Team_Id := To_Unbounded_String(Value(Get_Named_Item(Attributes(Current_Element), "id")));
            -- if the first team is the same as the visitor team (if any), get the
            -- second team element
            if Local_Team_Id = Teams_Status(Visitor).Get_Id then
               -- get the second team xml element
               Current_Element := Next_Sibling(Current_Element);
               while Node_Type(Current_Element) /= Element_Node loop
                  Current_Element := Next_Sibling(Current_Element);
               end loop;
               Local_Team_Id := To_Unbounded_String(Value(Get_Named_Item(
                 Attributes(Current_Element), "id")));
            end if;
         end;
         Load_Team(Local, Local_Team_Id, Teams_Configuration, True);
      end if;
      -- if there is not the configuration fot Visitor get the default configuration
      -- that do not conflict with the Local configuration (if any)
      if Teams_Status(Visitor).Get_Id = Null_Unbounded_String then
         declare
            Current_Element : Node :=
              First_Child(DOM.Core.Documents.Get_Element(Teams_Config_Xml_Doc));
         begin
            -- get the first team xml element
            while Node_Type(Current_Element) /= Element_Node loop
               Current_Element := Next_Sibling(Current_Element);
            end loop;
            Visitor_Team_Id := To_Unbounded_String(Value(Get_Named_Item(Attributes(Current_Element), "id")));
            -- if the first team is the same as the local team (if any), get the
            -- second team element
            if Visitor_Team_Id = Teams_Status(Local).Get_Id then
               -- get the second team xml element
               Current_Element := Next_Sibling(Current_Element);
               while Node_Type(Current_Element) /= Element_Node loop
                  Current_Element := Next_Sibling(Current_Element);
               end loop;
               Visitor_Team_Id := To_Unbounded_String(Value(Get_Named_Item(
                 Attributes(Current_Element), "id")));
            end if;
         end;
         Load_Team(Visitor, Visitor_Team_Id, Teams_Configuration, True);
      end if;
   end Load_Teams_Configuration;

   procedure Init_Players_Statistics(Team_Id : in Team_Id_T) is
   begin
      -- set the current status of active player to Active and get the shirt
      -- number for each player
      for I in Player_Number_T'Range loop
         if Players_Status(Team_Id, I).Get_State = Active then
            Statistics.Players_Statistics(Team_Id, I).Current_Status := Active;
         end if;
         Statistics.Players_Statistics(Team_Id, I).Shirt_Number :=
           Players_Characteristics(Team_Id, I).Shirt_Number;
      end loop;
   end Init_Players_Statistics;
end Configuration;
