with Game_Control; use Game_Control;
with Motion;
with Position; use Position;
with Ball; use Ball;
with Player; use Player;
with Status.Manager;
with Utils;
with Field; use Field;
with Status; use Status;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Referee; use Referee;

Package body Run_Time_Tests is
   -- set these variables to True to enable Run-Time tests
   Execute_Interruption_Tests : constant Boolean := True;
   Execute_Substitution_Tests : Boolean := True;

   -- This sleep use polling because we want it to be completely independent
   -- and uncoupled from the status and protocol procedures, so no
   -- synchronization is made.
   procedure Sleep is
   begin
      while Game_State.Get.State /= Play or not Ball_Status.Get.Free loop
         delay 1.0;
      end loop;
   end Sleep;

   procedure Simulate_Throw(From: Coordinates_3D_T; To: Coordinates_3D_T;
                            Player: Identifier_T;
                            Log: Unbounded_String := Null_Unbounded_String) is
   begin
      Sleep;

      Utils.Logger.Add_Log("TEST: " & To_String(Log));

      Status.Manager.Manager.Throw
        (Player_Id       => Player,
         Player_Position => Coordinates_T(From),
         New_Ball_Status => Ball.Ball_Status_T'
           (Motion            =>
              Motion.Calculate_Ball_Motion(From, To, 10.0),
            Current_Pos_Index => 0,
            Possessor         => Player,
            Free              => True)
        );
      -- Sleeping for some time here any other test is delayed, so that
      -- it's possible to see what happens when the game restarts
      while Game_State.Get.State = Play loop delay 0.1;
      end loop;
   end Simulate_Throw;

   procedure Test_Interruption(Tested_Interruption: State_T) is
      From   : Coordinates_3D_T;
      To     : Coordinates_3D_T;
      Player : Identifier_T;
      Log    : Unbounded_String;
   begin
      case Tested_Interruption is
         when Kick_Off	=>
            Log := To_Unbounded_String
              ("GOAL made by B11 and consequent KICK-OFF");
            From := Coordinates_3D_T'(Goal_X(Left) + 3, Goal_Y_Max - 1, 0.0);
            To   := Coordinates_3D_T'(Goal_X(Left), Goal_Y_Max - 1, 0.0);
            Player := (B, 11);
            Simulate_Throw(From,To,Player,Log);

            Log := To_Unbounded_String
              ("AUTOGOAL made by B2 and consequent KICK-OFF");
            From := Coordinates_3D_T'(Goal_X(Right) - 3, Goal_Y_Max - 1, 0.0);
            To   := Coordinates_3D_T'(Goal_X(Right), Goal_Y_Max - 1, 0.0);
            Player := (B, 2);
            Simulate_Throw(From,To,Player,Log);
         when Throw_In	=>
            Log := To_Unbounded_String
              ("THROW-IN for team B at the top touchline, left side");
            From := Coordinates_3D_T'(Midfield_Line / 2, Pitch.Top + 5, 0.0);
            To   := Coordinates_3D_T'(Midfield_Line / 2, Pitch.Top - 1, 0.0);
            Player := (A, 2);
            Simulate_Throw(From,To,Player,Log);

            Log := To_Unbounded_String
              ("THROW-IN for team B at the top touchline, right side");
            From := Coordinates_3D_T'(Midfield_Line * 3 / 2, Pitch.Top + 5, 0.0);
            To   := Coordinates_3D_T'(Midfield_Line * 3 / 2, Pitch.Top - 1, 0.0);
            Simulate_Throw(From,To,Player,Log);

            Log := To_Unbounded_String
              ("THROW-IN for team B at the bottom touchline, left side");
            From := Coordinates_3D_T'(Midfield_Line / 2, Pitch.Bottom -5, 0.0);
            To   := Coordinates_3D_T'(Midfield_Line / 2, Pitch.Bottom + 1, 0.0);
            Simulate_Throw(From,To,Player,Log);

            Log := To_Unbounded_String
              ("THROW-IN for team B at the bottom touchline, right side");
            From := Coordinates_3D_T'(Midfield_Line * 3 / 2, Pitch.Bottom-5, 0.0);
            To   := Coordinates_3D_T'(Midfield_Line * 3 / 2, Pitch.Bottom+1, 0.0);
            Simulate_Throw(From,To,Player,Log);

         when Goal_Kick	=>
            Log := To_Unbounded_String("GOAL KICK for team A");
            Player := (B, 2);
            From := Coordinates_3D_T'(Pitch.Left + 5, Goal_Y_Min - 5, 0.0);
            To   := Coordinates_3D_T'(Pitch.Left - 1, Goal_Y_Min - 5, 0.0);
            Simulate_Throw(From, To, Player, Log);

            Log := To_Unbounded_String("GOAL KICK for team B");
            Player := (A, 2);
            From := Coordinates_3D_T'(Pitch.Right - 5, Goal_Y_Min - 5, 0.0);
            To   := Coordinates_3D_T'(Pitch.Right + 1, Goal_Y_Min - 5, 0.0);
            Simulate_Throw(From, To, Player, Log);

         when Corner_Kick =>
            Log := To_Unbounded_String("CORNER KICK for team B, top touchline");
            Player := (A, 2);
            From := Coordinates_3D_T'(Pitch.Left + 5, Goal_Y_Min - 5, 0.0);
            To   := Coordinates_3D_T'(Pitch.Left - 1, Goal_Y_Min - 5, 0.0);
            Simulate_Throw(From, To, Player, Log);

            Log := To_Unbounded_String("CORNER KICK for team B, bottom touchline");
            Player := (A, 2);
            From := Coordinates_3D_T'(Pitch.Left + 5, Goal_Y_Max + 5, 0.0);
            To   := Coordinates_3D_T'(Pitch.Left - 1, Goal_Y_Max + 5, 0.0);
            Simulate_Throw(From, To, Player, Log);

            Log := To_Unbounded_String("CORNER KICK for team A, top touchline");
            Player := (B, 2);
            From := Coordinates_3D_T'(Pitch.Right - 5, Goal_Y_Min - 5, 0.0);
            To   := Coordinates_3D_T'(Pitch.Right + 1, Goal_Y_Min - 5, 0.0);
            Simulate_Throw(From, To, Player, Log);

            Log := To_Unbounded_String("CORNER KICK for team A, bottom touchline");
            Player := (B, 2);
            From := Coordinates_3D_T'(Pitch.Right - 5, Goal_Y_Max + 5, 0.0);
            To   := Coordinates_3D_T'(Pitch.Right + 1, Goal_Y_Max + 5, 0.0);
            Simulate_Throw(From, To, Player, Log);
         when others => null;
      end case;

   end Test_Interruption;

   function Test_Substitution(Outgoing_Player : in Identifier_T;
                              Incoming_Player : in Identifier_T)
                              return Boolean is
      Substitution_Accepted : Boolean := False;
   begin
      Utils.Logger.Add_Log
        ("TEST: Substitution of " & Outgoing_Player.Team_Id'Img
         & Outgoing_Player.Shirt_Number'Img & " with "
         & Incoming_Player.Team_Id'Img & Incoming_Player.Shirt_Number'Img);
      Referee.Referee_Task.Substitution
        (Substitution_Request => Substitution_Request_T'
           (Outgoing_Id => Outgoing_Player,
            Incoming_Id => Incoming_Player),
         Success              => Substitution_Accepted);
      if Substitution_Accepted then
         Utils.Logger.Add_Log("TEST: accepted");
      else
         Utils.Logger.Add_Log("TEST: refused");
      end if;
      return Substitution_Accepted;
   end;

   task body Tester is
   begin
      if Execute_Interruption_Tests or Execute_Substitution_Tests then
         Utils.Logger.Add_Log("Tester running");
      end if;

      if Execute_Substitution_Tests then
         while Game_State.Get.State /= Play loop delay 1.0; end loop;
         if not Test_Substitution((A,3), (A,14))   -- 1st
           or not Test_Substitution((A,7), (A,18)) -- 2nd
           or Test_Substitution((A,5), (B,6))      -- invalid
         then
            Utils.Logger.Add_Log("SUBSTITUTION TEST FAILED");
         end if;
      end if;

      delay 5.0;

      if Execute_Interruption_Tests then
         for Interruption in Kick_Off..Corner_Kick loop

            if Game_State.Get.State = Half_Time_Break then
               delay 20.0;
               if Execute_Substitution_Tests then
                  if not Test_Substitution((A,9), (A,20)) -- 3rd
                    or Test_Substitution((A,5), (A,16))	 -- 4th, should be refused
                  then
                     Utils.Logger.Add_Log("SUBSTITUTION TEST FAILED");
                  end if;
               end if;
               Execute_Substitution_Tests := False; -- they've been already done
            end if;

            if Game_State.Get.State = Play then
               Test_Interruption(Interruption);
            end if;
            delay 1.0;

         end loop;
      end if;

      delay 5.0;

      if Execute_Substitution_Tests then
         while Game_State.Get.State /= Half_Time_Break loop delay 1.0; end loop;
         delay 20.0;
         if not Test_Substitution((A,9), (A,20)) -- 3rd
           or Test_Substitution((A,5), (A,16))	 -- 4th, should be refused
         then
            Utils.Logger.Add_Log("SUBSTITUTION TEST FAILED");
         end if;
      end if;

   end Tester;

end Run_Time_Tests;
